#!/usr/bin/python

"""
analyze.py: Analyzes itu_channel/wavelength/crtt that was stored in a file

-------------------------------------------------------------------------------
Copyright (C) 2017 Peter Jansweijer, Henk Peek

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>
-------------------------------------------------------------------------------

Usage:
  analyze.py     <name>
  analyze.py     -h | --help

  <name>     if name is a filename and the file contains the crtt
             measurements then this file is analyzed and plotted
             if the name is a directory then the files in this
             directory are analyzed and the result is output in a
             file "result.txt"

Options:
  -h --help    Show this screen.
  -t           tolerance [ps], skip points outside fitted line, default: "200"
  -i           use ITU channels instead of wavelength

"""

import os
import sys
import scipy
from scipy.optimize import curve_fit
import numpy
import matplotlib.pyplot as plt

import pdb

# Add parent directory (containing 'lib') to the module search path
lib_path = (os.path.dirname(os.path.abspath(__file__)))
lib_path = os.path.join(lib_path,"..")
sys.path.insert(0,lib_path)

# private imports:
import lib.itu_conversions as itu_conv
import lib.wrt_sfppt015sc as tunable

############################################################################


def file_to_dict(filename):
  """
  Retrieve the measurement file
   
  filename -- source file from which to retrieve data.
    
  returns: True, <type 'dict'>
              key        : value
              itu_channel: [wavelength, crtt, temperature, SFP ch_numb]
           False
              "Not an In Situ Alpha measurements file"
  """

  # create an empty dictionairy
  crtt_data = {}
  data_file = open(filename, "r")

  line = data_file.readline()
  if line.strip() != "#In Situ Alpha measurements":
    Exception("file_to_dict: Not an In Situ Alpha measurements file.")
    data_file.close()
    return (False, crtt_data)

  line = data_file.readline()	# Read date
  line = data_file.readline()	# Read time
  line = data_file.readline()	# Read comment

  # create an empty lists for fields:
  # sfp module channel, ITU channel, ITU wavelength [nm], crtt

  while 1:
    line = data_file.readline()
    if line[:len("#")] != "#":  # Skip lines that are commented out
      line_lst = line.split(",")
      if len(line_lst) < 4:
        break

      # Values of interest on one line are:
      # channel, ITU channel, ITU wavelength [nm], crtt [ps], drxm, drxs
      # note that there may be multiple measurment lines (link restarts) for a give set
      # of {channel, ITU channel, ITU wavelength}
      lst_meas = []
      lst_meas.append(float(line_lst[2]))	        # Wavelength

      # try if there is already an entry for the ITU channel on this line
      # if so then read the already existing list of crtt measurements
      try:
        existing_crtt_list = crtt_data[float(line_lst[1])][1]
      except:
        existing_crtt_list = []

      try:
        existing_temp_list = crtt_data[float(line_lst[1])][1]
      except:
        existing_temp_list = []

      existing_crtt_list.append(float(line_lst[3]))    # append the crtt measurement of the current line
      existing_temp_list.append(float(line_lst[6]))    # append the temperature measurement of the current line
      lst_meas.append(existing_crtt_list)              # append the list of all crtt measurments
      lst_meas.append(existing_temp_list)              # append the list of all temperature measurments
      lst_meas.append(int(line_lst[0]))		       # append sfp module channel number
    
      crtt_data.update({float(line_lst[1]): lst_meas})	# itu_channel: [wavelength, [list_of_crtt], SFP ch_numb]

  data_file.close()

  return (True, crtt_data)

############################################################################


def dict_to_narray(crtt_data):
  """
  converts the data that is stored in crtt_data <type 'dict'> into
  <type 'numpy.ndarray'>
   
  returns: <type 'numpy.ndarray'>
  """

  crtt_array = {}
  lst_channel = []
  lst_itu_channel = []
  lst_itu_wavelength = []
  lst_crtt = []
  lst_temp = []

  for key_itu_ch, value_itu_ch in sorted(crtt_data.items()):
    lst_itu_channel.append(key_itu_ch)
    lst_itu_wavelength.append(value_itu_ch[0])

    #lst_crtt.append(value_itu_ch[1])
    # crtt measurements is a list of one (or more for multiple link
    # restarts) crtt measurments.
    # calculate the mean value of the all crtt measurements and
    # temperature measurements
    crtt_meas = numpy.array(value_itu_ch[1])
    crtt_mean = numpy.mean(crtt_meas)
    lst_crtt.append(crtt_mean)
    temp_meas = numpy.array(value_itu_ch[2])
    temp_mean = numpy.mean(temp_meas)
    lst_temp.append(temp_mean)
    
    lst_channel.append(value_itu_ch[2])
    
  crtt_array["itu_channel"]=scipy.array(lst_itu_channel)
  crtt_array["itu_wavelength"]=scipy.array(lst_itu_wavelength)
  crtt_array["crtt"]=scipy.array(lst_crtt)
  crtt_array["temp"]=scipy.array(lst_temp)
  crtt_array["channel"]=scipy.array(lst_channel)

  return (crtt_array)

############################################################################


def spec_alpha(alpha):
  """
  returns the value of alpha as it is used by the SPEC
  (see also White Rabbit Calibration procudeure:
  http://www.ohwr.org/attachments/4092/WR_Calibration-v1.1-20151109.pdf
  equation 11.

  alpha         -- <float> alpha value

  returns:
  alpha         -- <float> alpha according to SPEC 

  """
  return(2**40 * (((alpha+1)/(alpha+2))-0.5))

############################################################################

def calc_alpha_l1(lambda_1, lambda_2, crtt_lambda_1, crtt_lambda_2, fixed_lambda):
  """
  calc_alpha_l1 calculates the alpha factor for wavelength lambda l1.
  For this caluculation we need two cable round trip times that were measured
  using lamdba l1 and lambda l2 while both used a common fixed_lambda for the
  return channel.

  lambda_1, lambda_2, fixed_lambda  -- <float> lambda's in [m]
  crtt_l1, crtt_l2                  -- <float> cable round trip times measured
                                               at lambda_1 and lambda_2

  returns:
  alpha_lambda_1                    -- <float>

  """

  delta_lambda_1 = lambda_1 - fixed_lambda
  delta_lambda_2 = lambda_2 - fixed_lambda
  crtt_diff      = crtt_lambda_1 - crtt_lambda_2
  lambda_diff    = lambda_1 - lambda_2

  alpha_lambda_1 = (2 * delta_lambda_1 * crtt_diff)/((crtt_lambda_1 * lambda_diff) - (crtt_diff * delta_lambda_1))
  return (alpha_lambda_1)

############################################################################

def calc_alpha(l, crtt_l, tangent, fixed_lambda):
  """
  calc_alpha calculates the alpha factor for the wavelength lambda l.
  It takes the cable rount trip time that was measured at this lambda
  (crtt_l) while using fixed_lambda for the return path.
  The tangent is the slope of variation in time over wavelength.

  l                 -- <float> lambda [m]
  crtt_l            -- <int>   cable round trip time measured at lambda l
                               using fixed_lambda on the return channel [ps]
  tangent              -- <float> [ps/nm]
  fixed_lambda      -- <float> fixed return channel lambda [m]

  returns:
  alpha             -- <float> alpha value for lambda l when using fixed_lambda
                               as return channel

  """

  delta_l = (l - fixed_lambda)
  alpha = (2* tangent * delta_l )/(crtt_l - tangent * delta_l )

  return(alpha)

############################################################################


def line(x, a, b):
  return a * x + b

############################################################################
#
# If run from commandline, we can test the library
#
"""
Usage:
  analyze.py     <name>
  analyze.py     -h | --help

  <name>     if name is a filename and the file contains the crtt
             measurements then this file is analyzed and plotted
             if the name is a directory then the files in this
             directory are analyzed and the result is output in a
             file "result.txt"

Options:
  -h --help    Show this screen.
  -t           tolerance [ps], skip points outside fitted line, default: "200"
  -i           use ITU channels instead of wavelength

"""
if __name__ == "__main__":

  import argparse
  parser = argparse.ArgumentParser()
  parser.add_argument("name", help="file or directory name containing crtt measurements")
  parser.add_argument("-tolerance", default=200)
  parser.add_argument("-i", action='store_true')
  args = parser.parse_args()

  name = args.name
  tolerance = int(args.tolerance)
  use_itu_channels = args.i
  fixed_itu_channel = 20

  isfile = False
  files = []
  # Keep tangent and temperature values in an array in case multiple
  # files are processed.
  tangent_array = numpy.array([])
  temperature_array = numpy.array([])

  if os.path.exists(name) == True:
    if os.path.isdir(name) == True:
      print(name+": is a directory")

      # Any result files already there will be removed
      if os.path.isfile(os.path.join(name,"result.txt")) == True:
        os.remove(os.path.join(name, "result.txt"))

      # Get a file list (exclude files with extentions!)
      for file_name in os.listdir(name):
        if len(file_name.split('.')) <= 1:
          files.append(os.path.join(name, file_name))

      # (re) open result.txt file
      
      result_file = open(os.path.join(name, "result.txt"),"w")
      result_file.write("file: outlier-count, tangent, mean-temperature\n")

    elif os.path.isfile(name) == True:
      isfile = True
      print(name+": is a file")
      files.append(name)

    else:
     print("unexpected error")
     sys.exit()
  else:
    print(name+": does not exist")
    sys.exit()

  #print(files)

  for insitu_file in sorted(files):
    try:
      measurementfile, crtt_data = file_to_dict(insitu_file)
    except:
      print (insitu_file + ": wrong file type")
      continue

    if measurementfile == True:
      print(insitu_file)
      crtt_array = dict_to_narray(crtt_data)
      #pdb.set_trace()

      if use_itu_channels:
        x = crtt_array["itu_channel"]
      else:
        # Note that values are converted to [nm] instead of [m]
        # the curve fit function seems not to be able to handle otherwise
        x = 1e9*crtt_array["itu_wavelength"]
      y = crtt_array["crtt"]
      t = crtt_array["temp"]

      # create an array of length (x) elements with a tolerance value
      tol = numpy.zeros(len(x))
      tol.fill(tolerance)

      try:
        popt, pcov = curve_fit(line, x, y)
        # define the best fit line based on array (still with outliers):
        ly=[]
        for i in x:
          ly.append(line(i,popt[0],popt[1]))
        
        # loop over all array elements and list the indexes of the
        # outliers based on best fit line + tolerance
        outlier_idx = []
        outlier_cnt = 0
        
        for i in range(len(x)):
          if abs(ly[i] - y[i]) >= tolerance:
            outlier_idx.append(i)
            outlier_cnt = outlier_cnt + 1
            #print(x[i],"is an outlier")
        
        # create "clean" arrays where the outliers are removed based on the indexes found above
        x_clean = x  # in [nm]!
        y_clean = y
        t_clean = t
        for i in sorted(outlier_idx, reverse=True):
          x_clean = numpy.delete(x_clean, i)
          y_clean = numpy.delete(y_clean, i)
          t_clean = numpy.delete(t_clean, i)
        
        # Tackle the case when all datapoints are regarded as outlier due to narrow tolerance
        if len(x_clean) == 0:
          print("Scattered datapoints => all outliers!")
          print("Try to loosen the tolerance (for example -t 1000)")
          sys.exit()
        
        popt_clean, pcov_clean = curve_fit(line, x_clean, y_clean)
        tangent = popt_clean[0]  # in [ps]/[nm]
        tangent_array = numpy.append(tangent_array,tangent)
        # for this particular tangent measurement, take the mean value of the 
        # temperatures during the measurement.
        temperature = t_clean.mean()
        temperature_array = numpy.append(temperature_array,temperature)

        # again define the best fit line, now based on clean arrays:
        ly_clean=[]
        alpha_clean=[]
        
        fixed_lambda = itu_conv.itu_2_wavelength(fixed_itu_channel)
        
        ### Test. Take ITU channel 11 as a refernce and calculate alpha for this
        # wavelength using a scan over all other lambda's and crtt measurments
        alpha_tst=[]
        tst_lambda_1 = crtt_array["itu_wavelength"][0] # ITU channel 11 is refernce
        tst_crtt_1   = crtt_array["crtt"][0]
        for i in range(1,len(x_clean)):
          l2 = x_clean[i]/1e9  # convert back to [m]
          crtt_l2 = y_clean[i]
          alpha_tst.append(calc_alpha_l1(tst_lambda_1, l2, tst_crtt_1, crtt_l2, fixed_lambda))
        
        # scan through all wavelengths and calculate the individual alpha's
        # using the tangent that was found from the fitted line through the
        # clean datapoints.
        # Note: that the tangent is in [ps]/[nm] for proper alpha calculation
        # we need [ps]/[m]
        for i in x_clean:  # scan through all wavelengths in [nm]!
          clean_y = line(i,popt_clean[0],popt_clean[1])
          ly_clean.append(clean_y)
          alpha_clean.append(calc_alpha(i/1e9, clean_y, tangent*1e9, fixed_lambda))
        
        fig = plt.figure("CRTT versus ITU Channel number")
        ax = fig.add_subplot(111)
        ax.set_ylabel('CRTT [ps]')
        ax.text(0.01, 0.95, str(insitu_file), transform=ax.transAxes)
        ax.text(0.01, 0.90, "tolerance: " + str(tolerance) + "; outliers: " + str(outlier_cnt), transform=ax.transAxes)
        ax.text(0.01, 0.85, "fit y = a * x + b", transform=ax.transAxes)
        ax.text(0.01, 0.80, "a:" + str(popt_clean[0]), transform=ax.transAxes)
        ax.text(0.01, 0.75, "b:" + str(popt_clean[1]), transform=ax.transAxes)
        
        if use_itu_channels:
          ax.set_xlabel('ITU Channel number')
          ax.text(0.01, 0.70, "tangent:" + str(tangent) + " [ps/ITU channel]", transform=ax.transAxes)
        else:
          ax.set_xlabel('Wavelenth [nm]')
          ax.text(0.01, 0.70, "tangent:" + str(tangent) + " [ps/nm]", transform=ax.transAxes)
        
        ax.plot(x, y)              # Original datapoints
        ax.plot(x, ly, ".")        # best fit line
        ax.plot(x, ly + tol, ".")  # best fit line with + tolerances
        ax.plot(x, ly - tol, ".")  # best fit line with - tolerances
        ax.plot(x_clean, y_clean)  # cleaned array in [nm]
        ax.plot(x_clean, ly_clean) # final clean fit line in [nm]
          
        fig.savefig(insitu_file+".png")
        
        if isfile: # Analyze single file => plot 
          fig_alpha = plt.figure("alpha as a function of wavelength")
          ax = fig_alpha.add_subplot(111)
          if use_itu_channels:
            ax.set_xlabel('ITU Channel number')
          else:
            ax.set_xlabel('Wavelenth [nm]')
          ax.set_ylabel('Alpha')
          ax.axhline(0, color='gray')
          ax.axvline(fixed_lambda*1e9, color='red')
          ax.plot(x_clean, alpha_clean)
        
          alpha_file = open(name+".out","w")
          alpha_file.write("file: " + insitu_file + "\n")
          alpha_file.write("outlier-count: " + str(outlier_cnt) + "\n")
          alpha_file.write("tangent: " + str(tangent)+ " [ps]/[nm]\n")
          alpha_file.write("============================\n")
          alpha_file.write("ITU_Channel, Lambda, alpha, aplha(SPEC format)\n")
          for i in range(len(x_clean)):
            wavelength = x_clean[i]
            itu_ch = itu_conv.wavelength_2_itu(wavelength/1e9)
            alpha_file.write(str(itu_ch) + ", " + str(wavelength) + ", " + str(alpha_clean[i]) + ", " + str (spec_alpha(alpha_clean[i])) + "\n")
          alpha_file.close()
        
          #pdb.set_trace()
        
          plt.draw()
          plt.show()
        
        else: # Analyze multiple files => output result file
          fig.clf() # clear figure (it was stored in a png file)
          result_file.write(insitu_file + ": " + str(outlier_cnt) + ", " + str(tangent) + ", " + str(temperature) + "\n")

      except:
        print("file probably only contians measurements for a single wavelength")
        continue

    #else: measurementfile == False => wrong file type

  # Finally, after scanning all files in a directory (i.e. not a single file)
  # => add statistics:
  if not(isfile):
    result_file.write("\n")
    result_file.write("tolerance applied: " + str(tolerance) + "\n")
    result_file.write("tangent:\n")
    result_file.write("Mean:  " + str(tangent_array.mean()) +"\n")
    result_file.write("StDev: " + str(tangent_array.std(ddof=1)) +"\n")


sys.exit()
