#!/usr/bin/python

"""
analyze.py: Analyzes itu_channel/wavelength/crtt that was stored in a file

-------------------------------------------------------------------------------
Copyright (C) 2017 Peter Jansweijer, Henk Peek

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>
-------------------------------------------------------------------------------

Usage:
  analyze.py     <name>
  analyze.py     -h | --help

  <name>     if name is a filename and the file contains the crtt
             measurements then this file is analyzed and plotted
             if the name is a directory then the files in this
             directory are analyzed and the result is output in a
             file "result.txt"

Options:
  -h --help    Show this screen.
  -t           tolerance [ps], skip points outside fitted line, default: "200"
  -i           use ITU channels instead of wavelength

"""

import os
import sys
import scipy
from scipy.optimize import curve_fit
import numpy
import matplotlib.pyplot as plt

import pdb

# Add parent directory (containing 'lib') to the module search path
lib_path = (os.path.dirname(os.path.abspath(__file__)))
lib_path = os.path.join(lib_path,"..")
sys.path.insert(0,lib_path)

# private imports:
import lib.itu_conversions as itu_conv
import lib.wrt_sfppt015sc as tunable

############################################################################


def file_to_dict(filename):
  """
  Retrieve the measurement file
   
  filename -- source file from which to retrieve data.
    
  returns: True, <type 'dict'>
              key        : value
              itu_channel: [wavelength, crtt, temperature, SFP ch_numb]
           False
              "Not an In Situ Alpha measurements file"
  """

  # create an empty dictionairy
  crtt_data = {}
  data_file = open(filename, "r")

  line = data_file.readline()
  if line.strip() != "#In Situ Alpha measurements":
    Exception("file_to_dict: Not an In Situ Alpha measurements file.")
    data_file.close()
    return (False, crtt_data)

  line = data_file.readline()	# Read date
  line = data_file.readline()	# Read time
  line = data_file.readline()	# Read comment

  # create an empty lists for fields:
  # sfp module channel, ITU channel, ITU wavelength [nm], crtt

  while 1:
    line = data_file.readline()
    if line[:len("#")] != "#":  # Skip lines that are commented out
      line_lst = line.split(",")
      if len(line_lst) < 4:
        break

      # Values of interest on one line are:
      #   sfp_channel, ITU channel, ITU wavelength, crtt [ps], drxm, drxs,
      #   fiber-spool-temp, wr-slave-temp
      # note that there may be multiple measurment lines (link restarts) for a give set
      # of {channel, ITU channel, ITU wavelength}
      lst_meas = []
      lst_meas.append(float(line_lst[2]))	        # Wavelength

      # try if there is already an entry for the ITU channel on this line
      # if so then read the already existing list of crtt measurements
      try:
        existing_crtt_list = crtt_data[float(line_lst[1])][1]
      except:
        existing_crtt_list = []

      try:
        existing_temp_list = crtt_data[float(line_lst[1])][2]
      except:
        existing_temp_list = []

      existing_crtt_list.append(float(line_lst[3]))    # append the crtt measurement of the current line

      try:
        existing_temp_list.append(float(line_lst[6]))  # append the temperature measurement of the current line
      except:
        existing_temp_list.append(0.0)                 # case temp = None

      lst_meas.append(existing_crtt_list)              # append the list of all crtt measurments
      lst_meas.append(existing_temp_list)              # append the list of all temperature measurments
      lst_meas.append(int(line_lst[0]))		       # append sfp module channel number
    
      crtt_data.update({float(line_lst[1]): lst_meas})	# itu_channel: [wavelength, [list_of_crtt], [list_of_temp], SFP ch_numb]

  data_file.close()

  return (True, crtt_data)

############################################################################


def dict_to_narray(crtt_data):
  """
  converts the data that is stored in crtt_data <type 'dict'> into
  <type 'numpy.ndarray'>
   
  returns: <type 'numpy.ndarray'>
  """

  crtt_array = {}
  lst_channel = []
  lst_itu_channel = []
  lst_itu_wavelength = []
  lst_crtt = []
  lst_temp = []

  for key_itu_ch, value_itu_ch in sorted(crtt_data.items()):
    lst_itu_channel.append(key_itu_ch)
    lst_itu_wavelength.append(value_itu_ch[0])

    #lst_crtt.append(value_itu_ch[1])
    # crtt measurements is a list of one (or more for multiple link
    # restarts) crtt measurments.
    # calculate the mean value of the all crtt measurements and
    # temperature measurements
    crtt_meas = numpy.array(value_itu_ch[1])
    crtt_mean = numpy.mean(crtt_meas)
    lst_crtt.append(crtt_mean)
    temp_meas = numpy.array(value_itu_ch[2])
    temp_mean = numpy.mean(temp_meas)
    lst_temp.append(temp_mean)
    
    lst_channel.append(value_itu_ch[2])
    
  crtt_array["itu_channel"]=scipy.array(lst_itu_channel)
  crtt_array["itu_wavelength"]=scipy.array(lst_itu_wavelength)
  crtt_array["crtt"]=scipy.array(lst_crtt)
  crtt_array["temp"]=scipy.array(lst_temp)
  crtt_array["channel"]=scipy.array(lst_channel)

  return (crtt_array)

############################################################################


def spec_alpha(alpha):
  """
  returns the value of alpha as it is used by the SPEC
  (see also White Rabbit Calibration procudeure:
  http://www.ohwr.org/attachments/4092/WR_Calibration-v1.1-20151109.pdf
  equation 11.

  alpha         -- <float> alpha value

  returns:
  alpha         -- <float> alpha according to SPEC 

  """
  return(2**40 * (((alpha+1)/(alpha+2))-0.5))

############################################################################

def calc_alpha_l1(lambda_1, lambda_2, crtt_lambda_1, crtt_lambda_2, fixed_lambda):
  """
  calc_alpha_l1 calculates the alpha factor for wavelength lambda l1.
  For this caluculation we need two cable round trip times that were measured
  using lamdba l1 and lambda l2 while both used a common fixed_lambda for the
  return channel.

  lambda_1, lambda_2, fixed_lambda  -- <float> lambda's in [m]
  crtt_l1, crtt_l2                  -- <float> cable round trip times [ps]measured
                                               at lambda_1 and lambda_2

  returns:
  alpha_lambda_1                    -- <float>

  """

  delta_lambda_1 = lambda_1 - fixed_lambda
  delta_lambda_2 = lambda_2 - fixed_lambda
  crtt_diff      = crtt_lambda_1 - crtt_lambda_2
  lambda_diff    = lambda_1 - lambda_2

  alpha_lambda_1 = (2 * delta_lambda_1 * crtt_diff)/((crtt_lambda_1 * lambda_diff) - (crtt_diff * delta_lambda_1))
  return (alpha_lambda_1)

############################################################################

def calc_alpha_tangent(l, crtt_l, tangent, fixed_lambda):
  """
  calc_alpha_tangent calculates the alpha factor for the wavelength lambda l.
  It takes the cable rount trip time that was measured at this lambda
  (crtt_l) while using fixed_lambda for the return path.
  The tangent is the slope of variation in time over wavelength.

  l                 -- <float> lambda [m]
  crtt_l            -- <int>   cable round trip time measured at lambda l
                               using fixed_lambda on the return channel [ps]
  tangent           -- <float> [ps/nm]
  fixed_lambda      -- <float> fixed return channel lambda [m]

  returns:
  alpha             -- <float> alpha value for lambda l when using fixed_lambda
                               as return channel

  """

  delta_l = (l - fixed_lambda)
  alpha = (2* tangent * delta_l )/(crtt_l - tangent * delta_l )

  return(alpha)

############################################################################

def calc_alpha_sellmeier(l, fixed_lambda, A, B, C, D, E):
  """
  calc_alpha_sellmeier calculates the alpha factor for the wavelength lambda l
  while using fixed_lambda for the return path.
  It takes the Sellmeier cooefcients that came out of a cable rount trip time
  wavelength scan.

  l                 -- <float> lambda [m]
  fixed_lambda      -- <float> fixed return channel lambda [m]
  a..e              -- <float> Sellmeier terms

  returns:
  alpha             -- <float> alpha value for lambda l when using fixed_lambda
                               as return channel

  """

  bb = l**2 - fixed_lambda**2
  cc = l**-2 - fixed_lambda**-2
  dd = l**4 - fixed_lambda**4
  ee = l**-4 - fixed_lambda**-4

  alpha = 2* (B*bb + C*cc + D*dd + E*ee)/(A + B*fixed_lambda**2 + C*fixed_lambda**-2 + D*fixed_lambda**4 + E*fixed_lambda**-4)

  return(alpha)


############################################################################

def calc_alpha_error(alpha1, alpha2, crtt_fixed_lambda):
  """
  calc_alpha_error calculates the timing error between the PPS-ses of Master
  and Slave that will be the result of an "imperfect" alpha2 with respect to a
  "perfect" alpha1.
  The timing error scales linear with the round trip time (i.e. the length of
  your fiber)

  alpha1            -- <float> alpha to be considered the correct value
  alpha2            -- <float> alpha for which timing error needs to be calculated
  crtt_fixed_lambda -- <float> crtt value at the fixed wavelenght.
                               note that this is the virtual crtt for which
                               master2slave and slave2master wavelengths would be
                               equal. It can be estimated by extrapolating the crtt
                               array.

  returns:
  t_err             -- <float> PPS time error

  """

  t_err = ((alpha1-alpha2)*crtt_fixed_lambda)/4

  return(t_err)

############################################################################


def line(x, a, b):
  return a * x + b

############################################################################

def group_delay(l, A, B, C, D, E):
  """
  group_delay calculates the group delay for the wavelength lambda l based on
  the 5-term Sellmeier equation with coefficients A..E.
  
  5-term sellmeier group delay:
  tau(lambda) = A + B * lambda^2 + C * lambda^-2 + D * lambda^4 + E * lambda^-4

  A..E              -- <numpy.float64> Sellmeier coefficients
  l                 -- <float> lambda is in [nm].
                    -- Note! Convert to [m] before executing sellmeier fit

  returns:
  delay             -- <float> group delay value based on selm5 and l
  """

  # l = lambda is in [nm]. convert to [m] before executing sellmeier fit
  l = l/1e9
  sellmeier_5 = A + B * l**2 + C * l**-2 + D * l**4 + E * l**-4
  # and convert fitted value back to [nm]
  return sellmeier_5 * 1e9
  
############################################################################

# 5-term sellmeier 
# Dispersion coefficient (is the derivative of group_delay equation)
# D(lambda) = 2 * B * lambda - 2 * C * lambda^-3 + 4 * D * lambda^3 -4 * E * lambda^-5

#def dispersion_coefficient(l, b, c, d, e):
#  # l = lambda is in [nm]. convert to [m] before executing sellmeier fit
#  l = l/1e9
#  sellmeier_5 = 2 * b * l -2 * c * l**-3 + 4 * d * l**3 - 4 * e * l**-5
#  # and convert fitted value back to [nm]
#  return sellmeier_5 * 1e9

############################################################################

def alpha_err_writeln(alpha_err_file, l1, l2, l_fix, l_idx, A, B, C, D, E, crtt_fixed_lambda, alpha_3wl, time_err):

  """
  write a line to the alpha_err file in a pre formatted way.
  
  alpha_err_file    -- output file pointer
  l1,l2,l_fix       -- <float> 1, 2 and fixed lambda's in [nm].
  l_idx             -- <tuple> (idx1,idx2) lambda index into two dimensional array of alpha's and errors
  A..E              -- <float> sellmeier constants
  crtt_fixed_lambda -- <float> round trip time @ fixed lambda in [ps]
  alpha_3wl         -- <2D array of float> of alpha values determined via 3-wavelength method at lambda's l1 and l2
                       (axes: idx1 => l1, idx2 => l2)
  time_err          -- <2D array of float> of error values (axes: idx1 => l1, idx2 => l2)

  returns:
  """

  l1_str = "{0:.2f}".format(l1)
  l2_str = "{0:.2f}".format(l2)
  crtt_l1_str = "{0:.0f}".format(group_delay(l1, A, B, C, D, E))
  crtt_l2_str = "{0:.0f}".format(group_delay(l2, A, B, C, D, E))
  alpha_sel_str = "{0:.3e}".format(calc_alpha_sellmeier(l1/1e9, l_fix , A, B, C, D, E))
  alpha_3wl_str = "{0:.3e}".format(alpha_3wl[l_idx])
  alpha_spec_3wl_str = "{0:.0f}".format(spec_alpha(alpha_3wl[l_idx]))
  time_err_str = "{0:.3e}".format(time_err[l_idx])
  time_err_rel_str = "{0:.3e}".format(time_err[l_idx]/crtt_fixed_lambda)

  alpha_err_file.write(l1_str + "\t" + l2_str + "\t" + crtt_l1_str + "\t" + crtt_l2_str + "\t" + alpha_sel_str + "\t" + alpha_3wl_str + "\t" + alpha_spec_3wl_str + "\t\t\t" + time_err_str + "\t" + time_err_rel_str + "\n")

  return


############################################################################

def average_insitu_files(files):

  """
  Reads all insitu_files from a directory and calcualtes the average of all measurements.
  
  files             -- <list> of insitu_files to be analyzed

  returns:
  x                 -- <numpy.ndarray> array of wavelengths or ITU channels
  arr_y             -- <numpy.ndarray> mean value of all measured crtt values per wavelength
  arr_t             -- <numpy.ndarray> mean value of all measured temp values per wavelength
  """

  first = True
  lst_y = ([])
  lst_t = ([])
  
  for insitu_file in sorted(files):
    try:
      measurementfile, crtt_data = file_to_dict(insitu_file)
    except:
      print (insitu_file + ": wrong file type")
      continue
    if measurementfile == True:
      print(insitu_file)
      crtt_array = dict_to_narray(crtt_data)
      #pdb.set_trace()
  
      if first:
        first = False
        if use_itu_channels:
          x = crtt_array["itu_channel"]
        else:
          # Note that values are converted to [nm] instead of [m]
          # the curve fit function seems not to be able to handle otherwise
          x = 1e9*crtt_array["itu_wavelength"]
          #x = crtt_array["itu_wavelength"]

      lst_y.append(crtt_array["crtt"])
      lst_t.append(crtt_array["temp"])
    #else: measurementfile == False => wrong file type

  # Finally, after scanning all files in a directory (i.e. not a single file)
  # create numpy arrays from appended lists
  arr_y = numpy.array(lst_y)
  arr_t = numpy.array(lst_t)
  # Now each row contains a set of crtt measurements over a wavelength scan.
  # We want statistics over a single wavelength measurement, i.e. over the columns.
  # Therefore specify axis=0 for statistics calculations.
  
  #y_mean = arr_y.mean(axis=0)
  #y_std  = arr_y.std(axis=0,ddof=1)
  
  return(x, arr_y, arr_t)

############################################################################

def analyze_plot(insitu_file, analyse_single, x, y, name, tolerance, use_itu_channels, ref_name, fixed_itu_channel, tangent_array, temperature_array):

  """
  Reads all insitu_files from a directory and calculates the average of all measurements.
  
  insitu_file       -- insitu_files to be analyzed
  analyse_single    -- <boolean> if True as singel file is analyzed and all plots are displayed
  x                 -- <numpy.ndarray> array of wavelengths or ITU channels
  y                 -- <numpy.ndarray> measured crtt values per wavelength
  name              -- <string> file or directory name containing crtt measurements
  tolerance         -- <int> tolerance [ps], skip points outside fitted line
  use_itu_channels  -- <boolean> use ITU channels instead of wavelength
  ref_name          -- <string> file or directory name containing crtt reference measurements
  fixed_itu_channel -- <int> channel used for fixed Slave-to-Master communication
  tangent_array     -- <numpy.array> keep tangent values in an array in case multiple files are processed
  temperature_array -- <numpy.array> keep temperature values in an array in case multiple files are processed
  
  returns:
  tangent_array     -- <numpy.array> keep tangent values in an array in case multiple files are processed
  temperature_array -- <numpy.array> keep temperature values in an array in case multiple files are processed
  result_line       -- <string> summary line to be added to the result_file
  """

  # create an array of length (x) elements with a tolerance value
  tol = numpy.zeros(len(x))
  tol.fill(tolerance)
  result_line = ""
  
  try:
    
    # try to get control over the 8 ns spikes...
    # scan the crtt values, if value is less than 1 ns of the value of its successor then try and add 8 ns
    #for i in range(len(x)):
    #  if (i != 0) and (y[i] < y[i-1] - 1000):
    #    y[i] = y[i] + 8000
      
    #popt, pcov = curve_fit(line, x, y)
    popt, pcov = curve_fit(group_delay, x, y)

    # define the best fit line based on array (still with outliers):
    ly=[]
    for i in x:
      #ly.append(line(i,popt[0],popt[1]))
      ly.append(group_delay(i,popt[0],popt[1],popt[2],popt[3],popt[4]))
    ly=numpy.array(ly)

    # loop over all array elements and list the indexes of the
    # outliers based on best fit line + tolerance
    outlier_idx = []
    outlier_cnt = 0
   
    data_vs_fit_err = y - ly

    for i in range(len(x)):
      if abs(ly[i] - y[i]) >= tolerance:
        outlier_idx.append(i)
        outlier_cnt = outlier_cnt + 1
        #print(x[i],"is an outlier")
    
    # create "clean" arrays where the outliers are removed based on the indexes found above
    x_clean = x  # in [nm]!
    y_clean = y
    t_clean = t
    data_vs_fit_err_clean = data_vs_fit_err
    for i in sorted(outlier_idx, reverse=True):
      x_clean = numpy.delete(x_clean, i)
      y_clean = numpy.delete(y_clean, i)
      t_clean = numpy.delete(t_clean, i)
      data_vs_fit_err_clean = numpy.delete(data_vs_fit_err_clean, i)

    # Tackle the case when all datapoints are regarded as outlier due to narrow tolerance
    if len(x_clean) == 0:
      print("Scattered datapoints => all outliers!")
      print("Try to loosen the tolerance (for example -t 1000)")
      sys.exit()
    
    # first order linear fit:
    popt_clean, pcov_clean = curve_fit(line, x_clean, y_clean)
    #perr_clean = numpy.sqrt(numpy.diag(pcov_clean))
    #print("a =", popt_clean[0], "+/-", pcov_clean[0,0]**0.5)
    #print("b =", popt_clean[1], "+/-", pcov_clean[1,1]**0.5)
    
    tangent = popt_clean[0]  # in [ps]/[nm]
    tangent_stdev = pcov_clean[0,0]**0.5  # in [ps]/[nm]
    tangent_array = numpy.append(tangent_array,tangent)
    # for this particular tangent measurement, take the mean value of the 
    # temperatures during the measurement.
    temperature = t_clean.mean()
    temperature_max = t_clean.max()
    temperature_min = t_clean.min()
    temperature_array = numpy.append(temperature_array,temperature)
  
    # again define the best fit line, now based on clean arrays:
    ly_clean=[]
    ly_lin_clean=[]
    alpha_tangent=[]
    alpha_sellmeier=[]
    alpha_time_error=[]
    #alpha_linear=[]
  
    # get fixed_lambda and calculate the Cable Round Trip Time that would
    # result from a link where M2S and S2M would use the same "fixed" wavelength.
    fixed_lambda = itu_conv.itu_2_wavelength(fixed_itu_channel)
    crtt_fixed_lambda = group_delay(fixed_lambda*1e9,popt[0],popt[1],popt[2],popt[3],popt[4])
    
    # A clean set of datapoints for crtt results from ripping off all outliers.
    # Scan through all wavelengths and calculate the individual alpha's using:
    # 1) the tangent that was derived from a linear fit through a clean crtt scan
    # 2) the 5 Sellmeier terms derived from a 5-term Sellmeier fit through the clean crtt scan
    # 3) the 3 wavelength linear equations
    # Note 1: that the dispersion for the particular fiber length is in [ps]/[nm].
    # for proper alpha calculation we need [ps]/[m]
    # Note 2: fiber dispersion in [ps]/[nm]*[Km]. This can be calculated by dividing
    # the dispersion for the particular fiber length in [ps]/[nm] by the fiber length
    # in [Km].
  
    for i in x_clean:  # scan through all wavelengths in [nm]!
      lin_clean_y = line(i,popt_clean[0],popt_clean[1])
      clean_y = (group_delay(i,popt[0],popt[1],popt[2],popt[3],popt[4]))
      #dispersion = dispersion_coefficient(i,popt[1],popt[2],popt[3],popt[4])
      ly_lin_clean.append(lin_clean_y)  # linear
      ly_clean.append(clean_y)          # 5-t Sellmeier
  
      alpha_tan = calc_alpha_tangent(i/1e9, clean_y, tangent*1e9, fixed_lambda)
      alpha_sel = calc_alpha_sellmeier(i/1e9, fixed_lambda, popt[0],popt[1],popt[2],popt[3],popt[4])
  
      alpha_tangent.append(alpha_tan)
      #alpha_tangent.append(calc_alpha_tangent(i/1e9, clean_y, dispersion, fixed_lambda))
      alpha_sellmeier.append(alpha_sel)
      alpha_time_error.append(calc_alpha_error(alpha_sel, alpha_tan, crtt_fixed_lambda))
  
    # =====================================================
    # Plot CRTT versus ITU Channel number
    # =====================================================
    fig = plt.figure("CRTT versus ITU Channel number")
    ax = fig.add_subplot(111)
    ax.set_ylabel('CRTT [ps]')
    
    if use_itu_channels:
      ax.set_title("CRTT versus ITU Channel number")
      ax.set_xlabel('ITU Channel number')
      ax.text(0.01, 0.85, "1st order linear tangent: " + '{:5.3f}'.format(tangent) + " +/- " + '{:5.3f}'.format(tangent_stdev) + " [ps/ITU channel]", transform=ax.transAxes)
    else:
      ax.set_title("CRTT versus wavelength")
      ax.set_xlabel('Wavelenth [nm]')
      ax.text(0.01, 0.85, "1st order linear tangent: " + '{:5.3f}'.format(tangent) + " +/- " + '{:5.3f}'.format(tangent_stdev) + " [ps/nm]", transform=ax.transAxes)
    
    with_tolerances = False
    ax.text(0.01, 0.95, str(insitu_file), transform=ax.transAxes)
    ax.axvline(fixed_lambda*1e9, color='red')
    lns = ax.plot(x, y, color = 'blue', label='data')              # Original datapoints
    if with_tolerances:
      ax.text(0.01, 0.90, "tolerance: " + str(tolerance) + "; outliers: " + str(outlier_cnt), transform=ax.transAxes)
      #ax.text(0.01, 0.85, "fit y = a * x + b", transform=ax.transAxes)
      #ax.text(0.01, 0.80, "a:" + str(popt_clean[0]), transform=ax.transAxes)
      #ax.text(0.01, 0.75, "b:" + str(popt_clean[1]), transform=ax.transAxes)
      lns1 = ax.plot(x, ly, ".", color='gray', label='best fit 5-term Sellmeier')        # 5-term Sellmeier best fit
      ax.plot(x, ly + tol, ".", color='#c0c0c0', label='tolerances')  # 5-term Sellmeier best fit with + tolerances
      lns2 = ax.plot(x, ly - tol, ".", color='#c0c0c0')  # 5-term Sellmeier best fit with - tolerances
      lns3 = ax.plot(x_clean, y_clean, color = 'red', label='data, outliers removed')  # cleaned array in [nm]
      lns=lns+lns1+lns2+lns3
  
    lns1 = ax.plot(x_clean, ly_clean, color = 'green', label='5-term Sellmeier fit') # final clean 5th order sellmeier fit in [nm]
    lns2 = ax.plot(x_clean, ly_lin_clean, "-", color = 'blue', dashes=[5,5,5,5], label='linear fit') # final clean linear fit in [nm]
  
    ax2 = ax.twinx()
    ax2.set_ylabel('data versus fit err [ps]')

    # Calcualte standard deviation of data to fit error and format string
    stdev_str = "{0:.3f}".format(data_vs_fit_err_clean.std(ddof=1))
    lns3 = ax2.plot(x_clean, data_vs_fit_err_clean, color='grey', label='data vs. 5-term Sellmeier fit error\n   StDev: ' + stdev_str + '[ps]')
    lns=lns+lns1+lns2+lns3

    labels=[l.get_label() for l in lns]
    
    ax.legend(lns, labels, loc='lower right', fontsize='medium')
  
    fig.savefig(insitu_file+".png")
   
    if analyse_single: # => plot 
  
      # =====================================================
      # Plot alpha as a function of wavelength
      # =====================================================
      fig_alpha = plt.figure("alpha as a function of wavelength")
      ax = fig_alpha.add_subplot(111)
      ax.set_title("fiber Delay Coefficient (alpha) as a function of wavelength")
      if use_itu_channels:
        ax.set_xlabel('ITU Channel number')
      else:
        ax.set_xlabel('Wavelenth [nm]')
      ax.set_ylabel('Delay Coefficient (alpha)')
      #ax.set_ylabel('Alpha')
      #ax.text(0.01, 0.95, 'fiber delayCoefficient (alpha) derived using', transform=ax.transAxes)
      #ax.text(0.01, 0.90, 'linear and 5-term Sellmeier fit', transform=ax.transAxes)
      ax.axhline(0, color='gray')
      ax.axvline(fixed_lambda*1e9, color='red')
      ax.plot(x_clean, alpha_sellmeier, color='green', label='dirived from 5-term Sellmeier fit')
      ax.plot(x_clean, alpha_tangent, color='blue', dashes=[5,5,5,5], label='dirived from linear fit')
      #ax.plot(x_clean, alpha_linear, color='blue', dashes=[5,5,5,5], label='linear')
      ax.legend(loc='lower right', fontsize='medium')
      #pdb.set_trace()
      fig_alpha.subplots_adjust(left=0.17)

      # =====================================================
      # Plot Time error due to using linear instead of 5-term sellmeier
      # =====================================================
      fig_alpha_error = plt.figure("Sellmeier/Linear Error wavelength")
      ax = fig_alpha_error.add_subplot(111)
      ax.set_title("Time error due to using linear instead of 5-term sellmeier")
      if use_itu_channels:
        ax.set_xlabel('ITU Channel number')
      else:
        ax.set_xlabel('Wavelenth [nm]')
      ax.set_ylabel('time error [ps]')
      ax.text(0.01, 0.95, 'interpolated crtt when using fixed wavelengt for', transform=ax.transAxes)
      ax.text(0.01, 0.90, 'both mst2slv and slv2mst = '+ str(crtt_fixed_lambda) + '[ps]', transform=ax.transAxes)
      ax.axhline(0, color='gray')
      ax.axvline(fixed_lambda*1e9, color='red')
      arr_tangent_clean=numpy.array(alpha_tangent)
      arr_sellmeier_clean=numpy.array(alpha_sellmeier)
      ax.plot(x_clean, alpha_time_error, label='time error')
      ax.legend(loc='lower right', fontsize='medium')

      # =====================================================
      # 3D Plots
      # =====================================================
      from mpl_toolkits.mplot3d import axes3d
    
      # Here we are going to make the 3 wavelength error plot
      # l1 and l2 are 2D arrays of <float> lambda's in [nm]
      # l1 = array([[ 1568.77267399,  1567.95218619,  1567.13255619, ...,
      #               1530.33413987,  1529.55335714,  1528.77337073],
      #             [ 1568.77267399,  1567.95218619,  1567.13255619, ...,
      #               1530.33413987,  1529.55335714,  1528.77337073],
      #             ..., 
      #             [ 1568.77267399,  1567.95218619,  1567.13255619, ...,
      #               1530.33413987,  1529.55335714,  1528.77337073],
      #             [ 1568.77267399,  1567.95218619,  1567.13255619, ...,
      #               1530.33413987,  1529.55335714,  1528.77337073]])
      # l2 = array([[ 1568.77267399,  1568.77267399,  1568.77267399, ...,
      #               1568.77267399,  1568.77267399,  1568.77267399],
      #             [ 1567.95218619,  1567.95218619,  1567.95218619, ...,
      #               1567.95218619,  1567.95218619,  1567.95218619],
      #             ..., 
      #             [ 1529.55335714,  1529.55335714,  1529.55335714, ...,
      #               1529.55335714,  1529.55335714,  1529.55335714],
      #             [ 1528.77337073,  1528.77337073,  1528.77337073, ...,
      #               1528.77337073,  1528.77337073,  1528.77337073]])

      # crtt_l1 and crtt_l2 are 2D arrays of <float> crtt in [ps]
      l1,l2 = numpy.meshgrid(x_clean, x_clean)
      crtt_l1 = group_delay(l1,popt[0],popt[1],popt[2],popt[3],popt[4])
      crtt_l2 = group_delay(l2,popt[0],popt[1],popt[2],popt[3],popt[4])
      alpha_3wl = calc_alpha_l1(l1/1e9, l2/1e9, crtt_l1, crtt_l2, fixed_lambda)
      alpha_3wlsel = calc_alpha_sellmeier(l1/1e9, fixed_lambda, popt[0],popt[1],popt[2],popt[3],popt[4])
      alpha_err = alpha_3wlsel - alpha_3wl
      time_err = calc_alpha_error(alpha_3wlsel, alpha_3wl, crtt_fixed_lambda)
      time_err_relative = time_err/crtt_fixed_lambda
      
      #pdb.set_trace()
  
      """
      # =====================================================
      # Plot 3 wavelength alpha error
      # =====================================================
      fig_3wavelength_alpha_err = plt.figure("3 wavelength alpha error")
      ax = fig_3wavelength_alpha_err.add_subplot(111,projection='3d')
      ax.set_xlabel('l1 - fixed_lambda [nm]')
      ax.set_ylabel('l1 - l2 [nm]')
      ax.set_zlabel('alpha error')
      ax.plot_wireframe(l1-fixed_lambda*1e9, l1-l2, alpha_err)
  
      # =====================================================
      # Plot 3 wavelength time error (l1-lsm x-axis, l1-l2 y-axis, error z-axis)
      # =====================================================
      fig_3wavelength_time_err_offset = plt.figure("3 wavelength (relative displayed) time error for link with crtt(lambda_sm) =  "+ str(crtt_fixed_lambda) + "[ps]")
      ax = fig_3wavelength_time_err_offset.add_subplot(111,projection='3d')
      ax.set_xlabel('l1 - fixed_lambda [nm]')
      ax.set_ylabel('l1 - l2 [nm]')
      ax.set_zlabel('time error [ps]')
      ax.plot_wireframe(l1-fixed_lambda*1e9, l1-l2, time_err)
      """
  
      # =====================================================
      # Plot 3 wavelength time error (l1 on x-axis, l2 on y-axis, error z-axis)
      # =====================================================
      fig_3wavelength_time_err = plt.figure("3 wavelength time error for link with crtt(lambda_sm) =  "+ str(crtt_fixed_lambda) + "[ps]")
      ax = fig_3wavelength_time_err.add_subplot(111,projection='3d')
      ax.set_xlabel('l1 [nm]')
      ax.set_ylabel('l2 [nm]')
      ax.set_zlabel('time error [ps]')
      ax.plot_wireframe(l1, l2, time_err)
  
      """
      # =====================================================
      # Plot 3 wavelength relative time error (l1 on x-axis, l2 on y-axis, error z-axis)
      # =====================================================
      fig_3wavelength_time_err = plt.figure("3 wavelength relative time error")
      ax = fig_3wavelength_time_err.add_subplot(111,projection='3d')
      ax.set_xlabel('l1 [nm]')
      ax.set_ylabel('l2 [nm]')
      ax.set_zlabel('relative error [ps time error/ps link length]')
      ax.plot_wireframe(l1, l2, time_err_relative)
      """

      # =====================================================
      # Output (some) alpha values in file
      # =====================================================
      alpha_file = open(insitu_file+".out","w")
      alpha_file.write("file: " + insitu_file + "\n")
      alpha_file.write("reference file/directory: " + ref_name + "\n")
      alpha_file.write("outlier-count: " + str(outlier_cnt) + "\n")
      alpha_file.write("first order linear fit tangent: " + '{:5.3f}'.format(tangent) + " +/- " + '{:5.3f}'.format(tangent_stdev) + " [ps]/[nm]\n")
      alpha_file.write("============================\n")
      alpha_file.write("ITU_Channel, Lambda, alpha_sellmeier, aplha_sellmeier(SPEC format), alpha_tangent, alpha_sellmeier-alpha_tangent, temperature\n")
      for i in range(len(x_clean)):
        wavelength = x_clean[i]
        itu_ch = itu_conv.wavelength_2_itu(wavelength/1e9)
        alpha_file.write(str(itu_ch) + ", " + str(wavelength) + ", " + str(alpha_sellmeier[i]) + ", " + str (spec_alpha(alpha_sellmeier[i])) + ", " + str(alpha_tangent[i]) + ", " + str(alpha_sellmeier[i]- alpha_tangent[i]) + ", " + str(t_clean[i]) + "\n")
      alpha_file.close()
    
      # =====================================================
      # Output 3 wavelength alpha_error values in file
      # Also prepare a file for PPS delta delay measurement that
      #   contains lambda1 and alpha (in SPEC format).
      # =====================================================
      alpha_err_file = open(insitu_file+"_err.out","w")
      alpha_err_file.write("file: " + insitu_file + "\n")
      alpha_err_file.write("reference file/directory: " + ref_name + "\n")
      alpha_err_file.write("fixed_lambda: " + str(fixed_lambda) + "\n")
      alpha_err_file.write("Sellmeier: \n")
      alpha_err_file.write("A: " + str(popt[0]) + "\n")
      alpha_err_file.write("B: " + str(popt[1]) + "\n")
      alpha_err_file.write("C: " + str(popt[2]) + "\n")
      alpha_err_file.write("D: " + str(popt[3]) + "\n")
      alpha_err_file.write("E: " + str(popt[4]) + "\n")
  
      alpha_ddelay_file = open(insitu_file+"_ddelay.out","w")
      alpha_ddelay_file.write("# Proposed Delta Delay measurements\n")
      alpha_ddelay_file.write("# file: " + insitu_file + "\n")
      alpha_ddelay_file.write("# reference file/directory: " + ref_name + "\n")
      alpha_ddelay_file.write("# fixed_lambda: " + str(fixed_lambda) + " = ITU-ch: "+str(itu_conv.wavelength_2_itu(fixed_lambda))+"\n")

      alpha_err_file.write("min time_err: " + str(numpy.nanmin(time_err)) + "\n")
      l1_idx,l2_idx = numpy.where(time_err==numpy.nanmin(time_err))
      alpha_err_file.write("  @ l1: " + str(l2[l1_idx][0][0]) + "\n")
      alpha_err_file.write("  @ l2: " + str(l2[l2_idx][0][0]) + "\n")
    
      alpha_err_file.write("max time_err: " + str(numpy.nanmax(time_err)) + "\n")
      l1_idx,l2_idx = numpy.where(time_err==numpy.nanmax(time_err))
      alpha_err_file.write("  @ l1: " + str(l2[l1_idx][0][0]) + "\n")
      alpha_err_file.write("  @ l2: " + str(l2[l2_idx][0][0]) + "\n")
  
      rng = 1 # [ps]
      alpha_err_file.write("============================\n")
      alpha_err_file.write("time_err within range: " +str(rng) + " [ps]\n")
      alpha_ddelay_file.write("# ============================\n")
      alpha_ddelay_file.write("# time_err within range: " +str(rng) + " [ps]\n")
      #alpha_err_file.write("  @l1\tl2\n")
      rng_sel = numpy.argwhere(numpy.logical_and(time_err<rng,time_err>-rng))
      #pdb.set_trace()
      #for i in rng_sel:
      #  alpha_err_file.write(str(l2[i[0]][0]) + "\t" + str(l2[i[1]][0]) + "\n")

      alpha_err_file.write("l1[nm]\tl2[nm]\tcrtt_(l1)[ps]\tcrtt(l2)[ps]\talpha_sel(l1)\talpha_3wl(l1)\tSPEC_alpha_3wl(l1)\ttime_err[ps]\ttime_relative_err[t_err ps/link_length ps]\n")
      alpha_ddelay_file.write("# lambda-1[nm]\tITU-ch\tSPEC_alpha_3wl(l1)\tlambda-2[nm]\t\ttime_err[ps]\n")

      for i in rng_sel:
        lambda1 = l2[i[0]][0]
        lambda2 = l2[i[1]][0]
        lambda_idx = i[1],i[0]
        alpha_spec_3wl=spec_alpha(alpha_3wl[lambda_idx])
        itu_ch = itu_conv.wavelength_2_itu(lambda1/1e9)
        time_err_str = "{0:.3e}".format(time_err[lambda_idx]*1e-12)
      
        alpha_err_writeln(alpha_err_file, lambda1, lambda2, fixed_lambda, lambda_idx, popt[0], popt[1], popt[2], popt[3], popt[4], crtt_fixed_lambda, alpha_3wl, time_err)
        alpha_ddelay_file.write(str(lambda1)+", "+str(itu_ch)+", "+str(alpha_spec_3wl)+", "+str(lambda2)+", "+time_err_str+"\n")

      alpha_err_file.write("============================\n")
      alpha_err_file.write("4 lambda combinations along the full band:\n")
      alpha_ddelay_file.write("# ============================\n")
      alpha_ddelay_file.write("# 4 lambda combinations along the full band:\n")

      # fixed selection along the array:
      rng_sel=([[0,16],[0,32],[0,49],[16,0],[16,32],[16,49],[32,0],[32,16],[32,49],[49,0],[49,16],[49,32]])
      alpha_err_file.write("l1[nm]\tl2[nm]\tcrtt_(l1)[ps]\tcrtt(l2)[ps]\talpha_sel(l1)\talpha_3wl(l1)\tSPEC_alpha_3wl(l1)\ttime_err[ps]\ttime_relative_err[t_err ps/link_length ps]\n")
      alpha_ddelay_file.write("# lambda-1[nm]\tITU-ch\tSPEC_alpha_3wl(l1)\tlambda-2[nm]\t\ttime_err[ps]\n")
      for i in rng_sel:
        lambda1 = l2[i[0]][0]
        lambda2 = l2[i[1]][0]
        lambda_idx = i[1],i[0]
        alpha_spec_3wl=spec_alpha(alpha_3wl[lambda_idx])
        itu_ch = itu_conv.wavelength_2_itu(lambda1/1e9)
        time_err_str = "{0:.3e}".format(time_err[lambda_idx]*1e-12)

        alpha_err_writeln(alpha_err_file, lambda1, lambda2, fixed_lambda, lambda_idx, popt[0], popt[1], popt[2], popt[3], popt[4], crtt_fixed_lambda, alpha_3wl, time_err)
        alpha_ddelay_file.write(str(lambda1)+", "+str(itu_ch)+", "+str(alpha_spec_3wl)+", "+str(lambda2)+", "+time_err_str+"\n")

      alpha_err_file.write("============================\n")
      alpha_err_file.write("all possible combinations and their results:\n")
      alpha_err_file.write("l1[nm]\tl2[nm]\tcrtt_(l1)[ps]\tcrtt(l2)[ps]\talpha_sel(l1)\talpha_3wl(l1)\tSPEC_alpha_3wl(l1)\ttime_err[ps]\ttime_relative_err[t_err ps/link_length ps]\n")
      for l1_idx in range(len(x_clean)):
        for l2_idx in range(len(x_clean)):
          lambda1 = l1[l1_idx][l2_idx]
          lambda2 = l2[l1_idx][l2_idx]
          lambda_idx = l1_idx,l2_idx
          alpha_err_writeln(alpha_err_file, lambda1, lambda2, fixed_lambda, lambda_idx, popt[0], popt[1], popt[2], popt[3], popt[4], crtt_fixed_lambda,  alpha_3wl, time_err)

          #alpha_err_file.write(str(l1[l1_idx][l2_idx]) + "\t" + str(l2[l1_idx][l2_idx]) + "\t" + str(group_delay(l1[l1_idx][l2_idx],popt[0],popt[1],popt[2],popt[3],popt[4])) + "\t " + str(group_delay(l2[l1_idx][l2_idx],popt[0],popt[1],popt[2],popt[3],popt[4])) + "\t" + str(calc_alpha_sellmeier(l1[l1_idx][l2_idx]/1e9, fixed_lambda,popt[0],popt[1],popt[2],popt[3],popt[4])) + "\t " + str(alpha_3wl[l1_idx][l2_idx]) + "\t " + str(spec_alpha(alpha_3wl[l1_idx][l2_idx])) + "\t" + str(time_err[l1_idx][l2_idx]) + "\t" + str(time_err[l1_idx][l2_idx]/crtt_fixed_lambda) + "\n")
      alpha_err_file.close()
      alpha_ddelay_file.close()
  
  
      #plt.subplot_tool()
      plt.subplots_adjust(left=0.17)
      fig_alpha.savefig(insitu_file+"_alpha.png")
      plt.draw()
      plt.show()
    
    else: # Analyze multiple files => output result file
      fig.clf() # clear figure (it was stored in a png file)
      #numb_of_meas = numb_of_meas + 1
      #result_file.write(insitu_file + ": " + str(outlier_cnt) + ", " + str(tangent) + ", " + str(temperature_min) + ", " + str(temperature_max) + "\n")
      result_line = insitu_file + ": " + str(outlier_cnt) + ", " + str(tangent) + ", " + str(temperature_min) + ", " + str(temperature_max) + "\n"
  
  except:
    print("file probably only contains measurements for a single wavelength")
    #continue

  return(tangent_array, temperature_array, result_line)

###########################################################################
#
# If run from commandline, we can test the library
#
"""
Usage:
  analyze.py     <name>
  analyze.py     -h | --help

  <name>     if name is a filename and the file contains the crtt
             measurements then this file is analyzed and plotted
             if the name is a directory then the files in this
             directory are analyzed and the result is output in a
             file "result.txt"

Options:
  -h --help    Show this screen.
  -t           tolerance [ps], skip points outside fitted line, default: "200"
  -i           use ITU channels instead of wavelength
  -r           reference file with reference crtt measurements

"""
if __name__ == "__main__":

  import argparse
  parser = argparse.ArgumentParser()
  parser.add_argument("name", help="file or directory name containing crtt measurements")
  parser.add_argument("-tolerance", default=200)
  parser.add_argument("-i", action='store_true')
  parser.add_argument("-r", default='')
  args = parser.parse_args()

  name = args.name
  tolerance = int(args.tolerance)
  use_itu_channels = args.i
  ref_name = args.r
  fixed_itu_channel = 20

  analyse_single = False
  ref_analyse_single = False
  files = []
  ref_files = []
  # Keep tangent and temperature values in an array in case multiple
  # files are processed.
  tangent_array = numpy.array([])
  temperature_array = numpy.array([])

  # Figure out if the input is a single measurement file or a directory with files
  if os.path.exists(name) == True:
    if os.path.isdir(name) == True:
      print(name+": is a directory")

      # Any result files already there will be removed
      if os.path.isfile(os.path.join(name,"result.txt")) == True:
        os.remove(os.path.join(name, "result.txt"))

      # Get a file list (exclude files with extensions!)
      for file_name in os.listdir(name):
        if len(file_name.split('.')) <= 1:
          files.append(os.path.join(name, file_name))

      # (re) open result.txt file
      
      result_file = open(os.path.join(name, "result.txt"),"w")
      result_file.write("file: outlier-count, tangent, min-temperature, max-temperature\n")
      numb_of_meas = 0   # Keep track of the number of scans that is recorded in the result file

    elif os.path.isfile(name) == True:
      analyse_single = True
      print(name+": is a file")
      files.append(name)

    else:
     print("unexpected error")
     sys.exit()

  else:
    print(name+": does not exist")
    sys.exit()

  # Spot if there is a reference file or directory given.
  # If so, take the average values of the reference measurement files
  if ref_name != '' and os.path.exists(ref_name) == True:
    if os.path.isdir(ref_name) == True:
      print(ref_name+": is a reference directory")

      # Get a file list (exclude files with extentions!)
      for file_name in os.listdir(ref_name):
        if len(file_name.split('.')) <= 1:
          ref_files.append(os.path.join(ref_name, file_name))

    elif os.path.isfile(ref_name) == True:
      ref_analyse_single = True
      print(ref_name+": is a reference file")
      ref_files.append(ref_name)
    
    dump_x, arr_crtt,arr_temp = average_insitu_files(ref_files)
    # Each row contains a set of crtt measurements over a wavelength scan.
    # We want statistics over a single wavelength measurement, i.e. over the columns.
    # Therefore specify axis=0 for statistics calculations.
    ref_crtt_array = arr_crtt.mean(axis=0)
    ref_crtt_array_std  = arr_crtt.std(axis=0,ddof=1)
    # If a reference insitu scan was specified then this reference crtt
    # measurement (one measurement scan where the long fiber is replaced by a
    # short fiber) is subtracted from the long fiber measurment

  # Now that we have a set of (one or more) measurement files, analyse the files
  # one by one and save a plot on disk
  # In case there is only one file then create more plots and output.
  for insitu_file in sorted(files):
    try:
      measurementfile, crtt_data = file_to_dict(insitu_file)
    except:
      print (insitu_file + ": wrong file type")
      continue
  
    if measurementfile == True:
      print(insitu_file)
      crtt_array = dict_to_narray(crtt_data)
      #pdb.set_trace()
  
      if use_itu_channels:
        x = crtt_array["itu_channel"]
      else:
        # Note that values are converted to [nm] instead of [m]
        # the curve fit function seems not to be able to handle otherwise
        x = 1e9*crtt_array["itu_wavelength"]
        #x = crtt_array["itu_wavelength"]
  
      # If a reference insitu scan was specified then this reference crtt
      # measurement (one measurement scan where the long fiber is replaced by a
      # short fiber) is subtracted from the long fiber measurment
      if ref_name != '':
        y = crtt_array["crtt"] - ref_crtt_array
      else:
        y = crtt_array["crtt"] 
  
      t = crtt_array["temp"]
  
      tangent_array, temperature_array, result_line = analyze_plot(insitu_file, analyse_single, x, y, name, tolerance, use_itu_channels, ref_name, fixed_itu_channel, tangent_array, temperature_array)
      if not(analyse_single):
        numb_of_meas = numb_of_meas + 1
        result_file.write(result_line)

  # After scanning all files in a directory (i.e. not a single file)
  # => add statistics:
  if not(analyse_single):
    result_file.write("\n")
    result_file.write("tolerance applied: " + str(tolerance) + "\n")
    result_file.write("tangent:\n")
    result_file.write("Mean:  " + str(tangent_array.mean()) +"\n")
    result_file.write("StDev: " + str(tangent_array.std(ddof=1)) +"\n")
    result_file.write("StErr: " + str(tangent_array.std(ddof=1)/(numb_of_meas**0.5)) +"\n")
    result_file.write("Note: Individual files analysed. Statistics over individual results.\n")
    result_file.write("      Averaged output is average of individual crtt measurements treated as single measurement leading to average tangent.\n")

  # Finally average all measurement files and analyse the average is if it was a single file
  x, arr_crtt,arr_temp = average_insitu_files(files)
  # Each row contains a set of crtt measurements over a wavelength scan.
  # We want statistics over a single wavelength measurement, i.e. over the columns.
  # Therefore specify axis=0 for statistics calculations.
  crtt_array = arr_crtt.mean(axis=0)
  crtt_array_std  = arr_crtt.std(axis=0,ddof=1)
  temp_array = arr_temp.mean(axis=0)
  temp_array_std  = arr_temp.std(axis=0,ddof=1)

  # In case of single file analyse "name" has full path
  # else "name" is directory name and needs to be extended to a valid base name
  if analyse_single:
    insitu_file = name
  else:
    insitu_file = name + "/averaged"

  # Process the averaged measurements as a single measurement
  analyse_single = True
    
  if ref_name != '':
    y = crtt_array - ref_crtt_array
    y_stdev = numpy.sqrt(crtt_array_std**2 + ref_crtt_array_std**2)
  else:
    y = crtt_array

  t = temp_array

  tangent_array, temperature_array, result_line = analyze_plot(insitu_file, analyse_single, x, y, name, tolerance, use_itu_channels, ref_name, fixed_itu_channel, tangent_array, temperature_array)

sys.exit()
