#!/usr/bin/python

"""
pps_dif.py: Does multiple PPS difference measurements

-------------------------------------------------------------------------------
Copyright (C) 2016 Peter Jansweijer, Henk Peek
    
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>
-------------------------------------------------------------------------------

Usage:
  pps_diff.py     -name ddelay.out [-ms /dev/ttyUSB2] [-ss /dev/ttyUSB1] [-t <IP>] -tic <IP> [-o <dir>] [-a <number>] [-r <number>]
  pps_diff.py     -h | --help

Options:
  -h --help    Show this screen.
  -name        pointer to a "ddelay.out" file to be taken as input for measurements
  -ms          master serial port, default: "/dev/ttyUSB2"
  -ss          slave serial port, default: "/dev/ttyUSB1"
  -t <IP>      track fiber spool temperature (using PT100 and Digital Multimeter at IP address)
  -tic <IP>    Time Interval Counter IP number
  -o <dir>     optional directory for output file storage, default: "data/"
  -a <number>  number of delta delay measurements over which is averaged (default = 10)
  -r <number>  number of link restarts over which the delta_delay is averaged (default = 10)

"""

import os
import sys
import numpy
import scipy
import serial
import time
from multiprocessing import Process

import vxi11

import pdb

# Add parent directory (containing 'lib') to the module search path
lib_path = (os.path.dirname(os.path.abspath(__file__)))
lib_path = os.path.join(lib_path,"..")
sys.path.insert(0,lib_path)

#from lib.docopt import docopt
import lib.itu_conversions as itu_conv
import lib.wrt_sfppt015sc as tunable
import lib.LeCroy8254 as DSO
import lib.Keysight_53230A as TIC

###############################################
# Found a neat python timout function on:
# https://stackoverflow.com/questions/366682/how-to-limit-execution-time-of-a-function-call-in-python
# drawback: function cannot return a value
def run_with_limited_time(func, args, kwargs, time):
    """Runs a function with time limit

    :param func: The function to run
    :param args: The functions args, given as tuple
    :param kwargs: The functions keywords, given as dict
    :param time: The time limit in seconds
    :return: True if the function ended successfully. False if it was terminated.
    """
    p = Process(target=func, args=args, kwargs=kwargs)
    p.start()
    p.join(time)
    if p.is_alive():
        p.terminate()
        return False

    return True


###############################################
# Found a neat python timout function on:
# https://stackoverflow.com/questions/492519/timeout-on-a-function-call
# default retuns None when tim eout occurs    
def timeout(func, args=(), kwargs={}, timeout_duration=1, default=None):
    import signal

    class TimeoutError(Exception):
        pass

    def handler(signum, frame):
        raise TimeoutError()

    # set the timeout handler
    signal.signal(signal.SIGALRM, handler) 
    signal.alarm(timeout_duration)
    try:
        result = func(*args, **kwargs)
    except TimeoutError as exc:
        result = default
    finally:
        signal.alarm(0)

    return result


###############################################
def wr2wrpc(ser, cmd, prompt="slv=>"):

  ser.write("\r")
  time.sleep (0.1)
  ser.flushInput()

  for i in range(len(cmd)):
    ser.write(cmd[i])
    time.sleep (0.1)
  time.sleep (0.5)
  ser.readline()	# Readback command
  
  print(prompt + cmd)

  return

###############################################
def sfp_ena(ser_slave):
  # bring the link down and up again by dis-/en-abeling the TX laser
  # Note that this forces the wr switch to re-sybc and bitslide but
  # not the wr devidce (SPEC or CLB) connected to it; Their wr links
  # is kept up
  wr2wrpc(ser_slave,"sfp ena 0\r","slv=>")
  time.sleep (1)
  wr2wrpc(ser_slave,"sfp ena 1\r","slv=>")
  time.sleep (1)

  return()
  
  
###############################################
def link_restart(ser_master,ser_slave, master_is_switch = True):
  
  wr2wrpc(ser_slave,"ptp stop\r","slv=>")
  time.sleep (1)
  wr2wrpc(ser_slave,"mode slave\r","slv=>")
  time.sleep (1)
  wr2wrpc(ser_slave,"ptp start\r","slv=>")
  time.sleep (1)

  if master_is_switch:
    # A switch hald.sh restart causes all links to be restarted.
    # Note that this forces the wr switch as well as the devices connected
    # (SPEC or CLB) to re-sync and bitslide.
    wr2wrpc(ser_master,"/etc/init.d/hald.sh restart\r","mst=>")
    time.sleep (5)
  else:
    wr2wrpc(ser_master,"ptp stop\r","mst=>")
    time.sleep (1)
    wr2wrpc(ser_master,"mode gm\r","mst=>")
    time.sleep (20)
  
    wr2wrpc(ser_master,"ptp start\r","mst=>")
    time.sleep (1)
  
  sfp_ena(ser_slave)

  return()
  
  
###############################################
def wait_for_track_phase(ser_slave):

  sync_phase = False
  track_phase = False

  while not sync_phase:		# First wait for "SYNC_PHASE" state
    stat = ser_slave.readline()	# Readback status line
    stat_lst = stat.split(' ')	# split on spaces

    #print("Waiting for SYNC_PHASE:")
    #print(stat_lst)

    if len(stat_lst) >= 27:
      print(stat_lst[6])
      if "SYNC_PHASE" in stat_lst[6]:
        sync_phase = True

  err_cnt = 0

  while not track_phase:	# then wait for "SYNC_PHASE" state
    stat = ser_slave.readline()	# Readback status line
    stat_lst = stat.split(' ')	# split on spaces

    #print("Waiting for TRACK_PHASE:")
    #print(stat_lst)

    
    if len(stat_lst) < 27:
      err_cnt = err_cnt + 1
      print("### No track_phase: attempt: " + str(err_cnt) + "\n")
      #print("### No track_phase, retry ptp stop, start...\n")

      # bring the link down and up again by dis-/en-abeling the TX laser
      # Note that this forces the wr switch to re-sybc and bitslide but
      # not the wr devidce (SPEC or CLB) connected to it; Their wr links
      # is kept up
      sfp_ena(ser_slave)

      if err_cnt == 3:
        print("### retry ptp stop, start...")
        # stop any pending ptp and restart
        wr2wrpc(ser_slave,"ptp stop\r","slv=>")
        print(ser_slave.readline())
        wr2wrpc(ser_slave,"ptp start\r","slv=>")
        print(ser_slave.readline())
        print(ser_slave.readline())
        err_cnt = 0 # new attempt re-start counting errors
        #break			# break if not in "stat" output modus
    else:
      err_cnt = 0
      print(stat_lst[6])
      if "TRACK_PHASE" in stat_lst[6]:
        track_phase = True

  return(track_phase)
  #return()
  
  
###############################################
def get_statusline(ser_slave):

  stat_lst = []
  while len(stat_lst) < 27:             # Keep reading until
    stat = ser_slave.readline()         # Readback valid status line
    stat_lst = stat.split(' ')          # split on spaces

  return(stat_lst)

############################################################################
def osc_init(scope, time_base = 50.0e-9):

  """
  Initialize the Oscilloscope for PPS time difference measurement on channel 3 and 4.

    scope      -- instance of python-vxi connected to the oscilloscope
    time_base  -- <float> time base, default 50 ns/div
  """

  #scope =  vxi11.Instrument("192.168.32.243")
  print(scope.ask("*IDN?"))
  # Returns 'LECROY,WAVERUNNER8254M-MS,LCRY4253N20398,8.1.0'

  # Use Channel 3 PPS SPEC-1
  # use Channel 4 PPS SPEC-2

  # A fixed trigger level on channel 3.
  scope.write("TRIG_SELECT EDGE,SR,C3")

  scope.write("C3:TRIG_LEVEL 1.4")
  scope.write("C3:Volt_DIV 0.5")
  scope.write("C3:COUPLING D50")
  scope.write("C3:OFFSET -1.5")
  scope.write("VBS 'app.Acquisition.C3.InterpolateType = \"sinxx\"'")

  scope.write("C4:TRIG_LEVEL 1.4")
  scope.write("C4:Volt_DIV 0.5")
  scope.write("C4:COUPLING D50")
  scope.write("C4:OFFSET -1.5")
  scope.write("VBS 'app.Acquisition.C4.InterpolateType = \"sinxx\"'")

  # Trigger in the centre of the screen; important for maximum estimations
  # forwarded to function average_edge_to_sfd
  scope.write("TRIG_DELAY 0 ns")
  scope.write("TIME_DIV "+str(time_base))  # set 50 ns/div
  #scope.write("REFERENCE_CLOCK EXTERNAL")  # set external refrence clock

  scope.write("VBS 'app.Measure.P1.MeasurementType=\"measure\"'")
  scope.write("VBS 'app.Measure.P1.Source1=\"C3\"'")
  scope.write("VBS 'app.Measure.P1.Source2=\"C4\"'")
  scope.write("VBS 'app.Measure.P1.ParamEngine=\"DeltaDelay\"'")
  scope.write("VBS 'app.Measure.P1.view=True'")
  scope.write("VBS 'app.Measure.StatsOn=True'")

  return

############################################################################

def tic_init(counter):

  """
  Initialize the Time Interval Counter for PPS time difference measurement on channel 1 and 2.

    counter    -- instance of python-vxi connected to the Time Interval Counter
  """

  #counter =  vxi11.Instrument("192.168.32.251")
  print(counter.ask("*IDN?"))
  # Returns 'Agilent Technologies,53230A,MY50001484,02.05-1519.666-1.19-4.15-127-155-35'

  # Put the device in a known state
  counter.write("*RST")
  counter.write("*CLS")

  # Setup Time Interval measurement from channel 1 -> 2
  counter.write("CONFigure:TINTerval (@1),(@2)")
  counter.write("SENSe:GATE:STARt:DELAy:SOURce IMMediate")
  counter.write("SENSE:TINTerval:GATE:SOURce IMMediate")
  counter.write("SENSE:TINTerval:GATE:POLarity POSitive")

  # Use external 10MHz reference
  counter.write("ROSCillator:SOURce EXTernal")
  # Use internal 10MHz reference
  #counter.write("ROSCillator:SOURce INTernal")

  # Disable timeout (otherwise we might loose measurements)
  counter.write("SYSTem:TIMeout INFinity")

  # Channel 1 setup:
  counter.write("INP1:COUP DC")
  counter.write("INP1:RANGE 5")
  counter.write("INP1:SLOPE POS")
  counter.write("INP1:LEVEL:AUTO OFF")
  counter.write("INP1:IMP 50")
  counter.write("INP1:LEV 1.4")

  # Channel 2 setup:
  counter.write("INP2:COUP DC")
  counter.write("INP2:RANGE 5")
  counter.write("INP2:SLOPE POS")
  counter.write("INP2:LEVEL:AUTO OFF")
  counter.write("INP2:IMP 50")
  counter.write("INP2:LEV 1.4")

  return

############################################################################
def delta_delay_dso(scope, num_avg = 1):

  """
  Performs a Delta Delay Measurement.

    scope      -- instance of python-vxi connected to the oscilloscope
    num_avg    -- <int> number of measurements to average
  """

  scope.write("TRIG_MODE STOP")
  scope.write("CLEAR_SWEEPS")
  scope.write("TRIG_MODE SINGLE")

  cnt = 1
  for num in range(0,num_avg):
    scope.write("ARM;WAIT")
    while ('1' not in scope.ask("*OPC?")):
      pass


  return

############################################################################
def spec_alpha(alpha):
  """
  returns the value of alpha as it is used by the SPEC
  (see also White Rabbit Calibration procedure:
  http://www.ohwr.org/attachments/4092/WR_Calibration-v1.1-20151109.pdf
  equation 11.

  alpha         -- <float> alpha value

  returns:
  alpha         -- <float> alpha according to SPEC 

  """
  return(2**40 * (((alpha+1)/(alpha+2))-0.5))


############################################################################
def file_to_array(filename):
  """
  Retrieve the delat delay file
   
  filename -- source file from which to retrieve data.
    
  returns: True, <type 'numpy.array'>
              [itu-ch, SPEC_alpha_3wl],
              [itu-ch, SPEC_alpha_3wl],
                   :
           False
              "Not a Delta Delay file"
  """

  # create empty list
  pairs=[]

  data_file = open(filename, "r")

  line = data_file.readline()
  if line.strip() != "# Proposed Delta Delay measurements":
    Exception("file_to_arrays: Not a Delta Delay measurements file.")
    data_file.close()
    return (False, alpha_spec_data)

  while 1:
    line = data_file.readline()
    if line[:len("#")] != "#":  # Skip lines that are commented out
      line_lst = line.split(",")
      if len(line_lst) < 3:
        break

      # Values of interest on one line are:
      #   # lambda-1[nm], ITU-ch, SPEC_alpha_3wl(lambda-1)
      pair = float(line_lst[1]),float(line_lst[2])
      #   # lambda-1[nm], ITU-ch, alpha_sel(lambda-1)
      #pair = float(line_lst[1]),int(round(spec_alpha(float(line_lst[2]))))
      pairs.append(pair)

  arr_pair=numpy.array(pairs)
 
  data_file.close()

  return (True, arr_pair)

###############################################
# Main
###############################################

"""
Usage:
  pps_diff.py     -name ddelay.out [-ms /dev/ttyUSB2] [-ss /dev/ttyUSB1] [-t <IP>] -tic <IP> [-o <dir>] [-a <number>] [-r <number>]
  pps_diff.py     -h | --help

Options:
  -h --help    Show this screen.
  -name        pointer to a "ddelay.out" file to be taken as input for measurements
  -ms          master serial port, default: "/dev/ttyUSB2"
  -ss          slave serial port, default: "/dev/ttyUSB1"
  -t <IP>      track fiber spool temperature (using PT100 and Digital Multimeter at IP address)
  -tic <IP>    Time Interval Counter IP number
  -o <dir>     optional directory for output file storage, default: "data/"
  -a <number>  number of delta delay measurements over which is averaged (default = 10)
  -r <number>  number of link restarts over which the delta_delay is averaged (default = 10)

"""

if __name__ == "__main__":
  
  #arguments = docopt(__doc__,version='White Rabbit controlled via serial port')
  import argparse
  parser = argparse.ArgumentParser()
  parser.add_argument("name", help="file containing ddelay measurements to perform")
  parser.add_argument("-mserial", default="/dev/ttyUSB2")
  parser.add_argument("-sserial", default="/dev/ttyUSB1")
  parser.add_argument("-t", help="track fiber spool temperature (using PT100 and Digital Multimeter at IP address)", default=None)
  #parser.add_argument("-dso", default="192.168.32.243") # LeCroy8254M default IP
  parser.add_argument("-tic", default="192.168.32.251") # Time Interval Counter Agilent 53230A
  parser.add_argument("-output_dir", default="data")
  parser.add_argument("-a", default=10)
  parser.add_argument("-r", default=10)
  args = parser.parse_args()
  name = args.name
  print("Used Delta Delay input file: ",name)
  print("Use Master Serial port: ",args.mserial)
  print("Use Slave Serial port: ",args.sserial)
  print("Fiber spool temperature track: ",args.t)
  #print("Oscilloscope: ",args.dso)
  print("Time Interval Counter: ",args.tic)
  print("Output directory: ",args.output_dir)
  print("measurement avg: ",args.a)
  print("link restarts: ",args.r)

  if os.path.exists(name) == True and os.path.isfile(name) == True:
    measure, pairs = file_to_array(name)
  else:
    print("Delat Delay input file not found")
    sys.exit()

  # add trailing slash if not present
  output_dir = os.path.join(args.output_dir,'')
  if os.path.exists(output_dir) != True:
    os.mkdir(output_dir)
    print("Output directory does not exist => created: "+output_dir)

  # Some constants (for now)
  # scan from itu channel start to stop and skip the channel that is used
  # for the return channel. "itu_skip_width" can be made larger than 1
  # in case the dwdm filters are wider than just the return channel

  #fixed_delay_mst=int(328182)  # Master (dTx + dRx) [ps]
  #fixed_delay_slv=int(327963)  # Slave  (dTx + dRx) [ps]
  # A reference measurement will be give the CRTT's for a system without a long fiber.
  # This reference measurement will be subtracted from the CRTT measurements for a long fiber.
  # In this way, any system typical hardware delays are cancelled.
  # Initialize all hardware delays to 0. 
  fixed_delay_mst=int(0)       # Master (dTx + dRx) [ps]
  fixed_delay_slv=int(0)       # Slave  (dTx + dRx) [ps]

  master_is_switch = False     # If the master is a WR switch then the restart sequence is different from CLB or SPEC
  master_has_tunable = True    # determines whether the master or the slave has the tunable SFP (due to sfp functions needed
                               # this only works when master is a CLB or SPEC)
  
  # Master SFP (tunable)
  params = tunable.get_parameters()
  sfp_module_vendor_id = params["vendor_id"]
  itu_channel_start = params["itu_channel_start"]
  itu_channel_stop = params["itu_channel_stop"]

  # Slave SFP (fixed)
  vendor_id_slv = "EOLS-1612-24205D"

  itu_channel_skip = 20
  itu_skip_width = 0.5
  ddelay_measurement = int(args.a)   # number of delta delay measurements over which is averaged
  restarts           = int(args.r)   # number of link restarts over which the delta_delay is averaged

  #scope =  vxi11.Instrument(args.dso) 
  #osc_init(scope, time_base = 1.0e-9)

  # Initialize Time Interval Counter (1-2)
  # Use Channel 1 PPS SPEC-1 (50 Ohm, Level 1.4)
  # Use Channel 2 PPS SPEC-2 (50 Ohm, Level 1.4)
  counter =  vxi11.Instrument(args.tic) 
  tic_init(counter)

  spool_temp = None       # if not used then default None
  if args.t != None:      # open and configure Digital Multimeter with attached PT100
    dmm =  vxi11.Instrument(args.t)
    print(dmm.ask("*IDN?"))
    # Returns 'Keysight Technologies,34465A,MY57501367,A.02.14-02.40-02.14-00.49-03-01'

    # Configure for 4 wire temperature measurement
    dmm.write("SAMPle:COUNt 1")
    dmm.write("UNIT:TEMPerature C")
    print("Fiber Spool temperature", float(dmm.ask("MEAS:TEMP? FRTD")))

  ser_slave = serial.Serial()

  ser_slave.port = args.sserial
  ser_slave.baudrate = 115200
  ser_slave.parity = serial.PARITY_NONE
  ser_slave.bytesize = serial.EIGHTBITS
  ser_slave.stopbits = serial.STOPBITS_ONE
  ser_slave.timeout = None
  #ser_slave.timeout = 600 	# timeout 10 minutes
  ser_slave.open()

  ser_master = serial.Serial()

  ser_master.port = args.mserial
  ser_master.baudrate = 115200
  ser_master.parity = serial.PARITY_NONE
  ser_master.bytesize = serial.EIGHTBITS
  ser_master.stopbits = serial.STOPBITS_ONE
  ser_master.timeout = None
  #ser_master.timeout = 600 	# timeout 10 minutes
  ser_master.open()

  if (master_has_tunable):
    ser_tunable = ser_master
    ser_tunable_str = "mst=>"
  else:
    ser_tunable = ser_slave
    ser_tunable_str = "slv=>"
  
  #wr2wrpc(ser_master,"mode gm\r","mst=>")
  #wr2wrpc(ser_slave,"mode slave\r","slv=>")
  time.sleep (20)

  timestamp = time.localtime()
  
  filename=output_dir+time.strftime(format("%y%m%d_%H_%M_%S"),timestamp)+"_DeltaDelay_meas"
  print("save delta delay measurements into file:",filename)

  data_file = open(filename,"w")
  data_file.write("# Delta Delay measurements\n")
  data_file.write("# date:"+time.strftime(format("%d %b %Y"),timestamp)+"\n")
  data_file.write("# time:"+time.strftime(format("%H:%M:%S"),timestamp)+"\n")
  data_file.write("# delta delay, StdDev, fiber-spool temp\n")

  for pair in pairs:
    ch = pair[0]
    alpha = int(round(pair[1]))

    # Set Master Tunable to the preferred ITU channel
    if ch < itu_channel_start or ch > itu_channel_stop:
      print("ITU channel out of range "+str(itu_channel_start)+" to " +str(itu_channel_stop))
      exit()
    elif (ch > (itu_channel_skip - itu_skip_width) and ch < (itu_channel_skip + itu_skip_width)):
      print("Chosen ITU channel "+str(ch)+" is too close to Slave-Master return Channel "+str(itu_channel_skip))
      exit()
    else:
      sfp_ch = tunable.sfp_channel(ch)
      frequency = itu_conv.itu_2_frequency(ch)
      wavelength = itu_conv.itu_2_wavelength(ch)
      print("select ITU channel",ch, "which is sfp channel number",sfp_ch)

      lambda_str = "{0:.2f}".format(wavelength*1e9)
      data_file.write("# Selected ITU channel for Master->Slave: "+str(ch)+" = "+str(lambda_str)+" [nm], Alpha (SPEC format): "+str(alpha)+"\n")

      # stop any pending ptp and or statistics output
      wr2wrpc(ser_master,"stat off\r","mst=>")
      print(ser_master.readline())
      wr2wrpc(ser_slave,"stat off\r","slv=>")
      print(ser_slave.readline())
      wr2wrpc(ser_slave,"ptp stop\r","slv=>")

      # Set Master dTx = fixed delay (dTx+dRx), dRx = 0, alpha = 0
      wr2wrpc(ser_tunable,"sfp erase\r",ser_tunable_str)
      wr2wrpc(ser_tunable,"sfp add "+sfp_module_vendor_id+" "+str(fixed_delay_mst)+" 0 0\r",ser_tunable_str)
      print(ser_tunable.readline())
      wr2wrpc(ser_tunable,"sfp match\r",ser_tunable_str)
      print(ser_tunable.readline())

      # Set Slave dTx = fixed delay (dTx+dRx), dRx = 0, alpha
      wr2wrpc(ser_slave,"sfp erase\r","slv=>")
      wr2wrpc(ser_slave,"sfp add "+vendor_id_slv+" "+str(fixed_delay_slv)+" 0 "+str(alpha)+"\r","slv=>")
      print(ser_slave.readline())
      wr2wrpc(ser_slave,"sfp match\r","slv=>")
      print(ser_slave.readline())

      wr2wrpc(ser_tunable,"sfp sel_page2\r",ser_tunable_str)

      wr2wrpc(ser_slave,"stat on\r","slv=>")
  
      # select wavelength
      wr2wrpc(ser_tunable,"sfp wr_ch " + str(sfp_ch) + "\r",ser_tunable_str)
      print(ser_tunable.readline())
      wr2wrpc(ser_tunable,"sfp rd_ch\r",ser_tunable_str)
      print(ser_tunable.readline())

      restart = 0
      while restart < restarts:
      #for restart in range(restarts):
        print("Link restart: ",restart+1)

        link_restart(ser_master,ser_slave, master_is_switch)
          
        track_phase = False

        while not track_phase:
          # Call "wait_for_track_phase(ser_slave)" with a timeout of 60 seconds
          # If no TRACK_PHASE within 60 seconds then try restart link!
          #track_phase = run_with_limited_time(wait_for_track_phase, (ser_slave, ), {}, 60.0)
          #if not track_phase:
          track_phase = timeout(wait_for_track_phase, (ser_slave, ), {}, 60)
          if track_phase == None:
            print ("### Timeout waiting for TRACK_PHASE, try link restart...")
            link_restart(ser_master,ser_slave, master_is_switch)
          
        try:
          time.sleep (5) # skip first few measurements after TRACK_PHASE
          tic_filename = ""
          
          # default use Time Interval Counter for PPS measurement
          # Uncomment this is you want to use a LeCroy DSO
          """
          delta_delay_dso(scope,ddelay_measurement)
  
          # Get the statistics "mean" and StdDev Values.
          mean_str = str(scope.ask("VBS? 'Return=app.Measure.P1.Statistics(\"mean\").Result.Value'"))
          sdev_str = str(scope.ask("VBS? 'Return=app.Measure.P1.Statistics(\"sdev\").Result.Value'"))
          # scope.ask("VBS?... returns something like: u'VBS 0' which is a unicode that
          # needs to be converted to string. Next, strip "VBS " and convert to a value
          # <int> or <float>
          mean=float(mean_str[4:])
          sdev=float(sdev_str[4:])
          """

          meas_dat, tic_filename = TIC.get_time_interval(counter, ddelay_measurement, output_dir)
          lst_value = []
          for v in meas_dat:
            value = scipy.float64(v.strip())
            lst_value.append(value)
          arr_value = numpy.array(lst_value)
            
          mean= arr_value.mean()
          sdev=arr_value.std(ddof=1)
          
          if args.t != None:      # open and configure Digital Multimeter with attached PT100
            spool_temp =float(dmm.ask("MEAS:TEMP? FRTD"))

          print("delta delay: "+str(mean)+" Std Deviation: "+str(sdev)+" fiber-spool temp: "+str(spool_temp))
          data_file.write(str(mean)+", "+str(sdev)+", "+str(spool_temp)+", " + tic_filename + "\n")

        except:
          print ("### exception during ddelay_measurements.")
          continue

        restart = restart + 1
       
  # end for ch in itu_chs:

  data_file.close()
 
  ser_slave.close()
