#!/usr/bin/python

"""
wr_serial.py: Scans the cable round trip time (crtt) versus the wavelength that
              is set for an SFP with a tunable laser using a serial interface
              connection to the WR console.

-------------------------------------------------------------------------------
Copyright (C) 2016 Peter Jansweijer, Henk Peek
    
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>
-------------------------------------------------------------------------------

Usage:
  wr_serial.py     [-o <dir>] [-m <number>]
  wr_serial.py     -h | --help

Options:
  -h --help    Show this screen.
  -sm          master serial port, default: "/dev/ttyUSB2"
  -ss          slave serial port, default: "/dev/ttyUSB1"
  -t           track fiber spool temperature (using Digital Multimeter and PT100), default: False
  -o <dir>     optional directory for output file storage, default: "data/"
  -m <number>  number of measurement cycles (default = 1 cyle is a sweep from
               start to stop lambda)

"""

import os
import sys
import numpy
import serial
import time
from multiprocessing import Process

import pdb

# Add parent directory (containing 'lib') to the module search path
lib_path = (os.path.dirname(os.path.abspath(__file__)))
lib_path = os.path.join(lib_path,"..")
sys.path.insert(0,lib_path)

#from lib.docopt import docopt
import lib.itu_conversions as itu_conv
import lib.wrt_sfppt015sc as tunable


###############################################
# Found a neat python timout function on:
# https://stackoverflow.com/questions/366682/how-to-limit-execution-time-of-a-function-call-in-python
# drawback: function cannot return a value
def run_with_limited_time(func, args, kwargs, time):
    """Runs a function with time limit

    :param func: The function to run
    :param args: The functions args, given as tuple
    :param kwargs: The functions keywords, given as dict
    :param time: The time limit in seconds
    :return: True if the function ended successfully. False if it was terminated.
    """
    p = Process(target=func, args=args, kwargs=kwargs)
    p.start()
    p.join(time)
    if p.is_alive():
        p.terminate()
        return False

    return True


###############################################
# Found a neat python timout function on:
# https://stackoverflow.com/questions/492519/timeout-on-a-function-call
# default retuns None when tim eout occurs    
def timeout(func, args=(), kwargs={}, timeout_duration=1, default=None):
    import signal

    class TimeoutError(Exception):
        pass

    def handler(signum, frame):
        raise TimeoutError()

    # set the timeout handler
    signal.signal(signal.SIGALRM, handler) 
    signal.alarm(timeout_duration)
    try:
        result = func(*args, **kwargs)
    except TimeoutError as exc:
        result = default
    finally:
        signal.alarm(0)

    return result


###############################################
def wr2wrpc(ser, cmd, prompt="slv=>"):

  ser.write("\r")
  time.sleep (0.1)
  ser.flushInput()

  for i in range(len(cmd)):
    ser.write(cmd[i])
    time.sleep (0.1)
  time.sleep (0.5)
  ser.readline()	# Readback command
  
  print(prompt + cmd)

  return

###############################################
def sfp_ena(ser_slave):
  # bring the link down and up again by dis-/en-abeling the TX laser
  # Note that this forces the wr switch to re-sybc and bitslide but
  # not the wr devidce (SPEC or CLB) connected to it; Their wr links
  # is kept up
  wr2wrpc(ser_slave,"sfp ena 0\r","slv=>")
  time.sleep (1)
  wr2wrpc(ser_slave,"sfp ena 1\r","slv=>")
  time.sleep (1)

  return()
  
  
###############################################
def link_restart(ser_master,ser_slave, master_is_switch = True):
  
  wr2wrpc(ser_slave,"ptp stop\r","slv=>")
  time.sleep (1)
  wr2wrpc(ser_slave,"ptp start\r","slv=>")
  time.sleep (1)

  if master_is_switch:
    # A switch hald.sh restart causes all links to be restarted.
    # Note that this forces the wr switch as well as the devices connected
    # (SPEC or CLB) to re-sync and bitslide.
    wr2wrpc(ser_master,"/etc/init.d/hald.sh restart\r","mst=>")
    time.sleep (5)
  else:
    wr2wrpc(ser_master,"ptp stop\r","mst=>")
    time.sleep (1)
    #wr2wrpc(ser_master,"mode gm\r","mst=>")
    #time.sleep (20)
  
    wr2wrpc(ser_master,"ptp start\r","mst=>")
    time.sleep (1)
  
  sfp_ena(ser_slave)

  return()
  
  
###############################################
def wait_for_track_phase(ser_slave):

  sync_phase = False
  track_phase = False
  crtt = 0

  while not sync_phase:		# First wait for "SYNC_PHASE" state
    stat = ser_slave.readline()	# Readback status line
    stat_lst = stat.split(' ')	# split on spaces

    #print("Waiting for SYNC_PHASE:")
    #print(stat_lst)

    if len(stat_lst) >= 27:
      print(stat_lst[6])
      if "SYNC_PHASE" in stat_lst[6]:
        sync_phase = True

  err_cnt = 0

  while not track_phase:	# then wait for "SYNC_PHASE" state
    stat = ser_slave.readline()	# Readback status line
    stat_lst = stat.split(' ')	# split on spaces

    #print("Waiting for TRACK_PHASE:")
    #print(stat_lst)

    
    if len(stat_lst) < 27:
      err_cnt = err_cnt + 1
      print("### No track_phase: attempt: " + str(err_cnt) + "\n")
      #print("### No track_phase, retry ptp stop, start...\n")

      # bring the link down and up again by dis-/en-abeling the TX laser
      # Note that this forces the wr switch to re-sybc and bitslide but
      # not the wr devidce (SPEC or CLB) connected to it; Their wr links
      # is kept up
      sfp_ena(ser_slave)

      if err_cnt == 3:
        print("### retry ptp stop, start...")
        # stop any pending ptp and restart
        wr2wrpc(ser_slave,"ptp stop\r","slv=>")
        print(ser_slave.readline())
        wr2wrpc(ser_slave,"ptp start\r","slv=>")
        print(ser_slave.readline())
        print(ser_slave.readline())
        err_cnt = 0 # new attempt re-start counting errors
        #break			# break if not in "stat" output modus
    else:
      err_cnt = 0
      print(stat_lst[6])
      #try:
      #  crtt_lst = stat_lst[crtt_lst_idx].split(":")
      if "TRACK_PHASE" in stat_lst[6]:
        track_phase = True
      #except:
      #  print ("### error occured while reading wr status.")
      #  continue

  return(track_phase)
  #return()
  
  
###############################################
def get_statusline(ser_slave):

  stat_lst = []
  while len(stat_lst) < 27:             # Keep reading until
    stat = ser_slave.readline()         # Readback valid status line
    stat_lst = stat.split(' ')          # split on spaces

  return(stat_lst)


###############################################
# Main
###############################################

"""
Usage:
  wr_serial.py     [-o <dir>] [-m <number>]
  wr_serial.py     -h | --help

Options:
  -h --help    Show this screen.
  -sm          master serial port, default: "/dev/ttyUSB2"
  -ss          slave serial port, default: "/dev/ttyUSB1"
  -t <IP>      track fiber spool temperature (using PT100 and Digital Multimeter at IP address)
  -o <dir>     optional directory for output file storage, default: "data/"
  -m <number>  number of measurement cycles (default = 1 cyle is a sweep from
               start to stop lambda)

"""

if __name__ == "__main__":
  
  #arguments = docopt(__doc__,version='White Rabbit controled via serial port')
  import argparse
  parser = argparse.ArgumentParser()
  parser.add_argument("-mserial", default="/dev/ttyUSB2")
  parser.add_argument("-sserial", default="/dev/ttyUSB1")
  parser.add_argument("-t", help="track fiber spool temperature (using PT100 and Digital Multimeter at IP address)", default=None)
  parser.add_argument("-output_dir", default="data")
  parser.add_argument("-measurements", default=1, type=int)
  args = parser.parse_args()
  print("Use Master Serial port: ",args.mserial)
  print("Use Slave Serial port: ",args.sserial)
  print("Fiber spool temperature track: ",args.t)
  print("Output directory: ",args.output_dir)
  print("Number of measuremnt cycles: ", args.measurements)

  # add trailing slash if not present
  output_dir = os.path.join(args.output_dir,'')
  if os.path.exists(output_dir) != True:
    os.mkdir(output_dir)
    print("Output directory does not exist => created: "+output_dir)

  measurements = args.measurements
  
  # Some constants (for now)
  # scan from itu channel start to stop and skip the channel that is used
  # for the return channel. "itu_skip_width" can be made larger than 1
  # in case the dwdm filters are wider than just the return channel
  # Step size of the scan is defined by "itu_channel_increment"

  #fixed_delay_mst=int(328182)  # Master (dTx + dRx) [ps]
  #fixed_delay_slv=int(327963)  # Slave  (dTx + dRx) [ps]
  # A reference measurement will be give the CRTT's for a system without a long fiber.
  # This reference measurement will be subtracted from the CRTT measurements for a long fiber.
  # In this way, any system typical hardware delays are cancelled.
  # Initialize all hardware delays to 0. 
  fixed_delay_mst=int(0)       # Master (dTx + dRx) [ps]
  fixed_delay_slv=int(0)       # Slave  (dTx + dRx) [ps]
  master_is_switch = False     # If the master is a WR switch then the restart sequence is different from CLB or SPEC
  master_has_tunable = True    # detrmines wether the master or the slave has the tunable SFP (due to sfp functions needed
                               # this only works when master is a CLB or SPEC)
  
  # Master SFP (tunable)
  params = tunable.get_parameters()
  sfp_module_vendor_id = params["vendor_id"]
  itu_channel_start = params["itu_channel_start"]
  itu_channel_stop = params["itu_channel_stop"]
  #itu_channel_stop = 11.5  ## use this one for testing (it only tunes itu-11 and itu-11.5)

  # Slave SFP (fixed)
  vendor_id_slv = "EOLS-1612-24205D"

  itu_channel_skip = 20
  itu_skip_width = 0.5
  #itu_channel_increment = params["itu_channel_spacing"]
  itu_channel_increment = 1
  crtt_measurement  = 10  # number of crtt measurements over which the crtt is averaged
  crtt_skip         = 5   # skip first 5 crtt measurments
  restarts          = 1   # number of link restarts over which the crtt is averaged


  spool_temp = None       # if not used then default None
  if args.t != None:      # open and configue Digital Multimeter with attached PT100
    import vxi11
    dmm =  vxi11.Instrument(args.t)
    print(dmm.ask("*IDN?"))
    # Returns 'Keysight Technologies,34465A,MY57501367,A.02.14-02.40-02.14-00.49-03-01'

    # Configure for 4 wire temperature measurement
    dmm.write("SAMPle:COUNt 1")
    dmm.write("UNIT:TEMPerature C")
    print("Fiber Spool temperature", float(dmm.ask("MEAS:TEMP? FRTD")))

  ser_slave = serial.Serial()

  ser_slave.port = args.sserial
  ser_slave.baudrate = 115200
  ser_slave.parity = serial.PARITY_NONE
  ser_slave.bytesize = serial.EIGHTBITS
  ser_slave.stopbits = serial.STOPBITS_ONE
  ser_slave.timeout = None
  #ser_slave.timeout = 600 	# timeout 10 minutes
  ser_slave.open()

  ser_master = serial.Serial()

  ser_master.port = args.mserial
  ser_master.baudrate = 115200
  ser_master.parity = serial.PARITY_NONE
  ser_master.bytesize = serial.EIGHTBITS
  ser_master.stopbits = serial.STOPBITS_ONE
  ser_master.timeout = None
  #ser_master.timeout = 600 	# timeout 10 minutes
  ser_master.open()

  if (master_has_tunable):
    ser_tunable = ser_master
    ser_tunable_str = "mst=>"
  else:
    ser_tunable = ser_slave
    ser_tunable_str = "slv=>"
  
  wr2wrpc(ser_master,"mode gm\r","mst=>")
  wr2wrpc(ser_slave,"mode slave\r","slv=>")
  time.sleep (20)

  # stop any pending ptp and or statistics output
  wr2wrpc(ser_master,"stat off\r","mst=>")
  print(ser_master.readline())
  wr2wrpc(ser_slave,"stat off\r","slv=>")
  print(ser_slave.readline())
  wr2wrpc(ser_slave,"ptp stop\r","slv=>")

  # Set Master dTx = fixed delay (dTx+dRx), dRx = 0, alpha = 0
  wr2wrpc(ser_tunable,"sfp erase\r",ser_tunable_str)
  wr2wrpc(ser_tunable,"sfp add "+sfp_module_vendor_id+" "+str(fixed_delay_mst)+" 0 0\r",ser_tunable_str)
  print(ser_tunable.readline())
  wr2wrpc(ser_tunable,"sfp match\r",ser_tunable_str)
  print(ser_tunable.readline())

  # Set Slave dTx = fixed delay (dTx+dRx), dRx = 0, alpha = 0
  wr2wrpc(ser_slave,"sfp erase\r","slv=>")
  wr2wrpc(ser_slave,"sfp add "+vendor_id_slv+" "+str(fixed_delay_slv)+" 0 0\r","slv=>")
  print(ser_slave.readline())
  wr2wrpc(ser_slave,"sfp match\r","slv=>")
  print(ser_slave.readline())

  wr2wrpc(ser_tunable,"sfp sel_page2\r",ser_tunable_str)

  wr2wrpc(ser_slave,"stat on\r","slv=>")

  while measurements > 0:
    measurements = measurements - 1

    timestamp = time.localtime()
  
    filename=output_dir+time.strftime(format("%y%m%d_%H_%M_%S"),timestamp)+"_insitu_alpha_scan"
    print("save insitu_alpha_scan into file:",filename)
    data_file = open(filename,"w")
  
    data_file.write("#In Situ Alpha measurements\n")
    data_file.write("#date:"+time.strftime(format("%d %b %Y"),timestamp)+"\n")
    data_file.write("#time:"+time.strftime(format("%H:%M:%S"),timestamp)+"\n")
    data_file.write("#sfp_channel, ITU channel, ITU wavelength, crtt [ps], drxm, drxs, fiber-spool-temp, wr-slave-temp\n")
  
    # scan through selected wavelengths
    for ch in numpy.arange(itu_channel_start,itu_channel_stop+0.5, itu_channel_increment):
  
      # Exclude the ITU channel that is used for the return channel
      if not(ch > (itu_channel_skip - itu_skip_width) and ch < (itu_channel_skip + itu_skip_width)):
  
        sfp_ch = tunable.sfp_channel(ch)
        frequency = itu_conv.itu_2_frequency(ch)
        wavelength = itu_conv.itu_2_wavelength(ch)
        print("select ITU channel",ch, "which is sfp channel number",sfp_ch)
  
        # select a wavelength
        wr2wrpc(ser_tunable,"sfp wr_ch " + str(sfp_ch) + "\r",ser_tunable_str)
        print(ser_tunable.readline())
        wr2wrpc(ser_tunable,"sfp rd_ch\r",ser_tunable_str)
        print(ser_tunable.readline())
  
        restart = 0
        while restart < restarts:
        #for restart in range(restarts):
          print("Link restart: ",restart+1)

          link_restart(ser_master,ser_slave, master_is_switch)
          
          track_phase = False

          while not track_phase:
            # Call "wait_for_track_phase(ser_slave)" with a timeout of 60 seconds
            # If no TRACK_PHASE within 60 seconds then try restart link!
            #track_phase = run_with_limited_time(wait_for_track_phase, (ser_slave, ), {}, 60.0)
            #if not track_phase:
            track_phase = timeout(wait_for_track_phase, (ser_slave, ), {}, 60)
            if track_phase == None:
              print ("### Timeout waiting for TRACK_PHASE, try link restart...")
              link_restart(ser_master,ser_slave, master_is_switch)
          
          try:
            fail = 0
            meas_number = 0
            crtt = []
            while meas_number < (crtt_skip + crtt_measurement):   # take mean value over cttt_measurement
              stat_lst = timeout(get_statusline, (ser_slave, ), {}, 60)
              if stat_lst == None:
                print ("### Timeout waiting status line...")
                link_restart(ser_master,ser_slave, master_is_switch)
              else:
                #stat_lst = []
                #while len(stat_lst) < 27:             # Keep reading until
                #  stat = ser_slave.readline()         # Readback valid status line
                #  stat_lst = stat.split(' ')          # split on spaces
  
                #pdb.set_trace()
                #if len(stat_lst) >= 27:
                  # note:
                  # [s for s in stat_lst if "crtt" in s]
                  # returns a list with one item, for example ['crtt:82466168']
                  # take item [0] from this list, split it over ":" and take
                  # item [1] from the resulting list
                  # format for 'temp:' is different. it has an extra space so
                  # the temp value is the next item op the stat_lst
                curr_crtt = int([s for s in stat_lst if "crtt" in s][0].split(":")[1])
                curr_drxm = int([s for s in stat_lst if "drxm" in s][0].split(":")[1])
                curr_drxs = int([s for s in stat_lst if "drxs" in s][0].split(":")[1])
                curr_temp = float(stat_lst[stat_lst.index('temp:')+1])
                
                crtt.append(curr_crtt)
                if meas_number == 0:                  # first measurment loads bitslide check valiables
                  check_drxm = curr_drxm
                  check_drxs = curr_drxs
                  meas_number = meas_number + 1
                  print(curr_crtt, curr_drxm, curr_drxs)
                else:                                 # next measurements compare bitslide check variables
                  # check for good measurment; criteria:
                  # 1) bitslides should all be equal (no hidden link resynchronizations)
                  # 2) crtt outliers bigger than 200 ps are not excepted (no hidden finetime jumps;
                  #    a worry since wrpc-v4.1 and PPSi huge PPS offset repair)
                  if check_drxm == curr_drxm and check_drxs == curr_drxs and max(crtt) - min(crtt) < 200 :
                    meas_number = meas_number + 1
                    print(curr_crtt, curr_drxm, curr_drxs)
                    # print(".",end='') # <= python 3
                  else:
                    print ("### bitslide changed or crtt(max-min) > 200 ps. Take next couple of measurements...")
                    meas_number = 0                   # wrong measurement (link resync?) start all over.
                    crtt = []
                    fail = fail + 1                   # count the number of times this failed
                    if fail > 3:                      # restart link when this passes a certain threshold
                      print ("### failed too many times => restart link")
                      link_restart(ser_master,ser_slave, master_is_switch)
                      fail = 0
  
            crtt_mean = numpy.mean(crtt[(crtt_skip + 1):])     # skip first measurements

            if args.t != None:      # open and configue Digital Multimeter with attached PT100
              spool_temp =float(dmm.ask("MEAS:TEMP? FRTD"))

            print("sfp_channel: ", sfp_ch ,"itu_channel: ", ch ,"crtt avarage over "+str(crtt_measurement)+" measurments: "+str(crtt_mean)+" fiber-spool temp: "+str(spool_temp))
            data_file.write(str(sfp_ch)+", "+str(ch)+", "+str(wavelength)+", "+str(crtt_mean)+", "+str(curr_drxm)+", "+str(curr_drxs)+", "+str(spool_temp)+", "+str(curr_temp)+"\n")

          except:
            print ("### exception during crtt_measurements.")
            continue

          restart = restart + 1
          
    data_file.close()
  
  ser_slave.close()
